import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import global_mean_pool
from torch_geometric.nn import GATConv, HeteroConv
from torch import Tensor
from typing import Tuple, Optional, List, Dict
import pytorch_lightning as pl

from collections import defaultdict


import warnings


from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.module_dict import ModuleDict
from torch_geometric.typing import EdgeType, NodeType
from torch_geometric.utils.hetero import check_add_self_loops
import itertools


def group(xs: List[Tensor], aggr: Optional[str]) -> Optional[Tensor]:
    if len(xs) == 0:
        return None
    elif aggr is None:
        return torch.stack(xs, dim=1)
    elif len(xs) == 1:
        return xs[0]
    elif aggr == "cat":
        return torch.cat(xs, dim=-1)
    elif aggr == "myCat":
        return torch.cat(xs, dim=1)
    else:
        out = torch.stack(xs, dim=0)
        out = getattr(torch, aggr)(out, dim=0)
        out = out[0] if isinstance(out, tuple) else out
        return out


class MinibatchDiscrimination(nn.Module):
    def __init__(
        self, in_features, out_features, intermediate_features=75, initalisation=0.05
    ):
        super(MinibatchDiscrimination, self).__init__()
        self.T = nn.Parameter(
            torch.randn(in_features, intermediate_features, out_features)
            * initalisation
        )
        self.in_features = in_features
        self.intermediate_features = intermediate_features
        self.out_features = out_features

    def forward(self, x):
        # Compute minibatch features
        M = x @ self.T.view(
            x.shape[1], -1
        )  # (batch_size, intermediate_features * out_features)

        M = M.view(
            x.shape[0], -1, self.T.shape[2]
        )  # (batch_size, intermediate_features, out_features)

        # Compute L1 distance across samples
        M_exp = M.unsqueeze(0)  # (1, batch_size, intermediate_features, out_features)
        M_diff = M_exp - M_exp.transpose(
            0, 1
        )  # (batch_size, batch_size, intermediate_features, out_features)
        distance = torch.abs(M_diff).sum(dim=2)  # Sum over intermediate features

        exp_distance = torch.exp(-distance)  # (batch_size, batch_size, out_features)

        # Sum over all samples except itself
        minibatch_features = exp_distance.sum(dim=1) - 1  # (batch_size, out_features)

        return minibatch_features


def squeeze_heterogenous_representation(rep):
    for key in rep:
        rep[key] = rep[key].squeeze()
        if rep[key].dim() == 2:
            rep[key] = torch.unsqueeze(
                rep[key], dim=1
            )  # squeezed 1,1 to nothing, need to go back to 1,
    return rep


def aggr_attention_at_node(rep, layer, aggrs):
    for key in rep:
        # Aggregate features across different methods (sum, mean, max)
        for idx, aggr in enumerate(aggrs):
            if idx == 0:
                if aggr == "sum":
                    total_aggre = torch.sum(rep[key], dim=1)
                if aggr == "mean":
                    total_aggre = torch.sum(rep[key], dim=1) / rep[key].shape[1]
                if aggr == "max":
                    total_aggre = torch.amax(rep[key], dim=1)
            else:
                if aggr == "sum":
                    total_aggre = torch.cat(
                        (total_aggre, torch.sum(rep[key], dim=1)), dim=1
                    )
                if aggr == "mean":
                    total_aggre = torch.cat(
                        (total_aggre, torch.sum(rep[key], dim=1) / rep[key].shape[1]),
                        dim=1,
                    )
                if aggr == "max":
                    total_aggre = torch.cat(
                        (total_aggre, torch.amax(rep[key], dim=1)), dim=1
                    )

        rep[key] = total_aggre  # No in-place modification

    for key in rep:
        atten = torch.sigmoid(layer[f"{key}_attention"](rep[key]))

        # Instead of in-place multiplication, create a new tensor
        rep[key] = rep[key] * atten  # Avoid `rep[key] *= atten`

        rep[key] = layer[f"{key}_aggr"](rep[key])  # Pass through aggregation layer

    return rep


def heterogenous_layer_wrapper_w_cat_and_self_loop(
    layer,
    aggr_attention,
    selfloop_layer,
    x,
    edge_index,
    conditions,
    infeatures,
    aggrs=["sum", "mean", "max"],
):
    edge_index = {key: val for key, val in edge_index.items() if val.numel() != 0}

    outputs = layer(x, edge_index)

    outputs = aggr_attention_at_node(outputs, aggr_attention, aggrs)

    for node_type in outputs.keys():
        outputs[node_type] = F.leaky_relu(outputs[node_type])

    for node_type in x.keys():
        if (
            node_type in outputs
        ):  # if only some of the nodes of a node type are used as the end point of edges (as defined by edge_index) the un-used are output as 0. - dont want this, so take previous values
            mask = outputs[node_type].abs().sum(dim=1) == 0  # Find zero rows

            concatted_shape = torch.cat([outputs[node_type], x[node_type]], dim=1)

            outputs[node_type] = F.leaky_relu(
                selfloop_layer[node_type](concatted_shape)
            )

            outputs[node_type][mask] = x[node_type][mask]  # Restore original embeddings
        else:
            outputs[node_type] = x[node_type]  # Pass through unchanged

    return outputs


def print_heterogenous_representation(rep):
    for key in rep:
        print(f"{key}:   {rep[key].shape}     (sum: {torch.sum(rep[key])})")
    print("\n")


class HeteroGenerator(pl.LightningModule):
    def __init__(
        self,
        track_targets_dims,
        track_conditions_dims,
        hidden_channels,
        track_latent_dims,
        edge_conditions_dims,
        conditonless=False,
    ):
        super(HeteroGenerator, self).__init__()

        dropout_rate = 0.2
        self.dropout = nn.Dropout(p=dropout_rate)

        # Conditions
        C_track = track_conditions_dims
        if conditonless:
            C_track = 0
        self.conditonless = conditonless

        # Targets
        T_track = track_targets_dims

        # Latents
        L_track = track_latent_dims

        ### ### ### ### ###

        self.particle_types = ["11", "13", "211", "321", "2212"]
        self.edge_types = list(itertools.product(self.particle_types, repeat=2))

        self.communication_dims = 100

        self.selfloop_layers = nn.ModuleList()
        self.communication_layers = nn.ModuleList()

        # self.N_layers = 3

        # for N_layer in range(self.N_layers):

        #     layer = {}
        #     for particle_type in particle_types:
        #         if N_layer == 0:
        #             layer[particle_type] = nn.Linear(L_track + C_track, hidden_channels)
        #         else:
        #             layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
        #     layer = nn.ModuleDict(layer)
        #     self.selfloop_layers.append(layer)

        #     layer_dict = {}
        #     for edge_type in self.edge_types:
        #         layer_dict[edge_type[0], "to", edge_type[1]] = GATConv(
        #             self.communication_dims,
        #             self.communication_dims,
        #             add_self_loops=False,
        #             edge_dim=edge_conditions_dims,
        #         )
        #     layer = HeteroConv(
        #         layer_dict,
        #         aggr="max",
        #     )
        #     self.communication_layers.append(layer)

        # layer = {}
        # for particle_type in particle_types:
        #     layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
        # layer = nn.ModuleDict(layer)
        # self.selfloop_layers.append(layer)

        # layer = {}
        # for particle_type in particle_types:
        #     layer[particle_type] = nn.Linear(hidden_channels, T_track)
        # layer = nn.ModuleDict(layer)
        # self.selfloop_layers.append(layer)

        self.communication_dims = hidden_channels

        self.selfloop_layers = nn.ModuleList()
        self.communication_layers = nn.ModuleList()

        self.communication_atten_layers = nn.ModuleList()
        self.communication_aggr_layers = nn.ModuleList()
        self.communication_compressor_layers = nn.ModuleList()
        self.x_compressor_layers = nn.ModuleList()

        self.N_layers = 3
        # self.N_layers = 4
        self.hidden_channels = hidden_channels

        for N_layer in range(self.N_layers):
            layer = {}
            for particle_type in self.particle_types:
                if N_layer == 0:
                    layer[particle_type] = nn.Linear(L_track + C_track, hidden_channels)
                else:
                    layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
            layer = nn.ModuleDict(layer)
            self.selfloop_layers.append(layer)

            layer_dict = {}
            for edge_type in self.edge_types:
                layer_dict[edge_type[0], "to", edge_type[1]] = GATConv(
                    self.communication_dims,
                    self.communication_dims,
                    add_self_loops=False,
                    edge_dim=edge_conditions_dims,
                )
            layer = HeteroConv(
                layer_dict,
                # aggr="mean", # was max,
                aggr="max",  # was max,
            )
            self.communication_layers.append(layer)

            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(
                    hidden_channels, int(hidden_channels / 2)
                )
            layer = nn.ModuleDict(layer)
            self.communication_compressor_layers.append(layer)
            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(
                    hidden_channels, int(hidden_channels / 2)
                )
            layer = nn.ModuleDict(layer)
            self.x_compressor_layers.append(layer)

            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
            layer = nn.ModuleDict(layer)
            self.communication_atten_layers.append(layer)

            # layer = {}
            # for particle_type in self.particle_types:
            #     layer[particle_type] = nn.Linear(hidden_channels * 2, hidden_channels)
            # layer = nn.ModuleDict(layer)
            # self.communication_aggr_layers.append(layer)

        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
        layer = nn.ModuleDict(layer)
        self.selfloop_layers.append(layer)

        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(hidden_channels, T_track)
        layer = nn.ModuleDict(layer)
        self.selfloop_layers.append(layer)
        self.T_track = T_track

    def forward(
        self,
        track_latent: Tensor,
        batch,  # contains all edge_index information
        offline_query_mode=False,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        x = {}
        for track_type in track_latent:
            x[track_type] = torch.cat(
                (batch[f"{track_type}_conditions"].x, track_latent[track_type]), dim=1
            )

        edge_index_dict = {}
        edge_attr_dict = {}
        for edge_type in self.edge_types:
            edge_index_dict[edge_type[0], "to", edge_type[1]] = batch[
                edge_type[0], "to", edge_type[1]
            ].edge_index  # ["edge_index"]
            edge_attr_dict[edge_type[0], "to", edge_type[1]] = batch[
                edge_type[0], "to", edge_type[1]
            ].edge_attr  # ["edge_attr"]

        # x_comms = {}

        # for N_layer in range(self.N_layers):

        #     for track_type in x:
        #         x[track_type] = F.leaky_relu(
        #             self.selfloop_layers[N_layer][track_type](x[track_type])
        #         )
        #         x_comms[track_type] = x[track_type][:, : self.communication_dims]
        #     x_comms = self.communication_layers[N_layer](x_comms, edge_index_dict, edge_attr_dict)
        #     for track_type in x_comms:
        #         x_comms[track_type] = F.leaky_relu(x_comms[track_type])
        #     for track_type in x:
        #         x[track_type] = x[track_type].clone()
        #         x[track_type][:, : self.communication_dims] = x_comms[track_type]

        # for track_type in x:
        #     x[track_type] = F.leaky_relu(
        #         self.selfloop_layers[-2][track_type](x[track_type])
        #     )

        # for track_type in x:
        #     x[track_type] = torch.tanh(
        #         self.selfloop_layers[-1][track_type](x[track_type])
        #     )

        # return x

        x_comms = {}

        for N_layer in range(self.N_layers):
            # x_in = {}
            for track_type in x:
                if x[track_type].shape[0] > 0:
                    x[track_type] = F.leaky_relu(
                        self.selfloop_layers[N_layer][track_type](x[track_type])
                    )
                else:
                    x[track_type] = torch.empty((0, self.hidden_channels)).to(
                        x[track_type].device
                    )
            #         x_in[track_type] = x[track_type]
            # x = x_in
            # print_heterogenous_representation(x)

            x_comms = self.communication_layers[N_layer](
                x, edge_index_dict, edge_attr_dict
            )
            for track_type in x_comms:
                if x_comms[track_type].shape[0] > 0:
                    x_comms[track_type] = F.leaky_relu(x_comms[track_type])
                    x[track_type] = torch.cat(
                        [
                            self.x_compressor_layers[N_layer][track_type](
                                x[track_type]
                            ),
                            self.communication_compressor_layers[N_layer][track_type](
                                x_comms[track_type]
                            ),
                        ],
                        dim=-1,
                    )
                    atten = self.communication_atten_layers[N_layer][track_type](
                        x[track_type]
                    )
                    x[track_type] = x[track_type] + torch.sigmoid(atten) * x[track_type]

        # print_heterogenous_representation(x)

        for track_type in x:
            if x[track_type].shape[0] > 0:
                x[track_type] = self.selfloop_layers[-1][track_type](x[track_type])
            else:
                x[track_type] = torch.empty((0, self.T_track)).to(x[track_type].device)
        # print_heterogenous_representation(x)
        # quit()
        return x


class HeteroDiscriminator(pl.LightningModule):
    def __init__(
        self,
        track_targets_dims,
        track_conditions_dims,
        hidden_channels,
        track_latent_dims,
        edge_conditions_dims,
        conditonless=False,
        add_MBD=False,
        add_physics_dims=False,
    ):
        super(HeteroDiscriminator, self).__init__()

        dropout_rate = 0.2
        self.dropout = nn.Dropout(p=dropout_rate)

        self.use_spectral_norm = False
        spec_norm = torch.nn.utils.spectral_norm

        def apply_spec_norm(layer):
            """Apply spectral normalization if enabled."""
            return spec_norm(layer) if self.use_spectral_norm else layer

        # Conditions
        C_track = track_conditions_dims

        if conditonless:
            C_track = 0
        self.conditonless = conditonless

        # Targets
        T_track = track_targets_dims
        if add_physics_dims:
            T_track += 1

        # Latents
        L_track = track_latent_dims

        self.L_track = L_track

        ### ### ### ### ###

        extra_dims = 0
        # extra_dims = 4

        self.particle_types = ["11", "13", "211", "321", "2212"]
        self.edge_types = list(itertools.product(self.particle_types, repeat=2))

        # self.selfloop_layers = nn.ModuleList()
        # self.communication_layers = nn.ModuleList()

        # self.communication_dims = 100

        # # self.N_layers = 3
        # self.N_layers = 2

        # for N_layer in range(self.N_layers):

        #     layer = {}
        #     for particle_type in particle_types:
        #         if N_layer == 0:
        #             layer[particle_type] = nn.Linear(
        #                 T_track + C_track + extra_dims, hidden_channels
        #             )
        #         else:
        #             layer[particle_type] = nn.Linear(
        #                 hidden_channels, hidden_channels
        #             )
        #     layer = nn.ModuleDict(layer)
        #     self.selfloop_layers.append(layer)

        #     layer_dict = {}
        #     for edge_type in self.edge_types:
        #         layer_dict[edge_type[0], "to", edge_type[1]] = GATConv(
        #             self.communication_dims,
        #             self.communication_dims,
        #             add_self_loops=False,
        #             edge_dim=edge_conditions_dims,
        #         )
        #     layer = HeteroConv(
        #         layer_dict,
        #         aggr="max",
        #     )
        #     self.communication_layers.append(layer)

        # # layer = {}
        # # for particle_type in particle_types:
        # #     layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
        # # layer = nn.ModuleDict(layer)
        # # self.selfloop_layers.append(layer)

        # # layer_dict = {}
        # # for edge_type in self.edge_types:
        # #     layer_dict[edge_type[0], "to", edge_type[1]] = GATConv(
        # #         self.communication_dims,
        # #         self.communication_dims,
        # #         add_self_loops=False,
        # #         edge_dim=edge_conditions_dims,
        # #     )
        # # layer = HeteroConv(
        # #     layer_dict,
        # #     aggr="max",
        # # )
        # # self.communication_layers.append(layer)

        # layer = {}
        # for particle_type in particle_types:
        #     layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
        # layer = nn.ModuleDict(layer)
        # self.selfloop_layers.append(layer)

        # self.add_MBD = add_MBD

        # if self.add_MBD:
        #     self.mbd_dims = 5
        #     mbd_layers = {}
        #     for particle_type in particle_types:
        #         mbd_layers[particle_type] = MinibatchDiscrimination(
        #             hidden_channels, self.mbd_dims, 5, initalisation=0.001
        #         )
        #     self.mbd_layer = nn.ModuleDict(mbd_layers)

        # self.output_layer = nn.Linear(hidden_channels, 1)

        # layer = {}
        # for particle_type in particle_types:
        #     if self.add_MBD:
        #         layer[particle_type] = nn.Linear(hidden_channels+self.mbd_dims, hidden_channels)
        #     else:
        #         layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
        # layer = nn.ModuleDict(layer)
        # self.selfloop_layers.append(layer)

        #### new
        self.communication_dims = hidden_channels

        self.selfloop_layers = nn.ModuleList()
        self.communication_layers = nn.ModuleList()

        self.communication_atten_layers = nn.ModuleList()
        self.communication_aggr_layers = nn.ModuleList()
        self.communication_compressor_layers = nn.ModuleList()
        self.x_compressor_layers = nn.ModuleList()

        self.N_layers = 3
        # self.N_layers = 4
        self.hidden_channels = hidden_channels

        for N_layer in range(self.N_layers):
            layer = {}
            for particle_type in self.particle_types:
                if N_layer == 0:
                    layer[particle_type] = nn.Linear(
                        T_track + C_track + extra_dims, hidden_channels
                    )
                else:
                    layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
            layer = nn.ModuleDict(layer)
            self.selfloop_layers.append(layer)

            layer_dict = {}
            for edge_type in self.edge_types:
                layer_dict[edge_type[0], "to", edge_type[1]] = GATConv(
                    self.communication_dims,
                    self.communication_dims,
                    add_self_loops=False,
                    edge_dim=edge_conditions_dims,
                )
            layer = HeteroConv(
                layer_dict,
                # aggr="mean", # was max,
                aggr="max",  # was max,
            )
            self.communication_layers.append(layer)

            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(
                    hidden_channels, int(hidden_channels / 2)
                )
            layer = nn.ModuleDict(layer)
            self.communication_compressor_layers.append(layer)
            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(
                    hidden_channels, int(hidden_channels / 2)
                )
            layer = nn.ModuleDict(layer)
            self.x_compressor_layers.append(layer)

            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
            layer = nn.ModuleDict(layer)
            self.communication_atten_layers.append(layer)

            # layer = {}
            # for particle_type in self.particle_types:
            #     layer[particle_type] = nn.Linear(hidden_channels * 2, hidden_channels)
            # layer = nn.ModuleDict(layer)
            # self.communication_aggr_layers.append(layer)

        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
        layer = nn.ModuleDict(layer)
        self.selfloop_layers.append(layer)

        self.add_MBD = add_MBD

        if self.add_MBD:
            self.mbd_dims = 5
            mbd_layers = {}
            for particle_type in self.particle_types:
                mbd_layers[particle_type] = MinibatchDiscrimination(
                    hidden_channels, self.mbd_dims, 5, initalisation=0.001
                )
            self.mbd_layer = nn.ModuleDict(mbd_layers)

        self.output_layer = nn.Linear(hidden_channels, 1)

        layer = {}
        for particle_type in self.particle_types:
            if self.add_MBD:
                layer[particle_type] = nn.Linear(
                    hidden_channels + self.mbd_dims, hidden_channels
                )
            else:
                layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
        layer = nn.ModuleDict(layer)
        self.selfloop_layers.append(layer)

    def forward(
        self,
        batch,  # contains all edge_index information
        logits=False,
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
        # device = track_conditions.x.device

        particle_types = ["11", "13", "211", "321", "2212"]

        x = {}
        for track_type in particle_types:
            x[track_type] = torch.cat(
                (batch[f"{track_type}_conditions"].x, batch[f"{track_type}"].x), dim=1
            ).to(torch.float32)

            # print(batch[f"{track_type}_conditions"].x)
            # print(batch[f"{track_type}"].x)
            # print('a', x[track_type].dtype)

        edge_index_dict = {}
        edge_attr_dict = {}
        for edge_type in self.edge_types:
            edge_index_dict[edge_type[0], "to", edge_type[1]] = batch[
                edge_type[0], "to", edge_type[1]
            ]["edge_index"]
            edge_attr_dict[edge_type[0], "to", edge_type[1]] = batch[
                edge_type[0], "to", edge_type[1]
            ]["edge_attr"]

        # for i in x:
        #     x[i] = x[i].to('cuda:5')
        # for i in edge_index_dict:
        #     edge_index_dict[i] = edge_index_dict[i].to('cuda:5')
        # for i in edge_attr_dict:
        #     edge_attr_dict[i] = edge_attr_dict[i].to('cuda:5')

        x_comms = {}

        for N_layer in range(self.N_layers):
            # x_in = {}
            for track_type in x:
                if x[track_type].shape[0] > 0:
                    x[track_type] = F.leaky_relu(
                        self.selfloop_layers[N_layer][track_type](x[track_type])
                    )
                else:
                    x[track_type] = torch.empty((0, self.hidden_channels)).to(
                        x[track_type].device
                    )
            #         x_in[track_type] = x[track_type]
            # x = x_in
            # print_heterogenous_representation(x)

            x_comms = self.communication_layers[N_layer](
                x, edge_index_dict, edge_attr_dict
            )
            for track_type in x_comms:
                if x_comms[track_type].shape[0] > 0:
                    x_comms[track_type] = F.leaky_relu(x_comms[track_type])
                    x[track_type] = torch.cat(
                        [
                            self.x_compressor_layers[N_layer][track_type](
                                x[track_type]
                            ),
                            self.communication_compressor_layers[N_layer][track_type](
                                x_comms[track_type]
                            ),
                        ],
                        dim=-1,
                    )
                    atten = self.communication_atten_layers[N_layer][track_type](
                        x[track_type]
                    )
                    x[track_type] = x[track_type] + torch.sigmoid(atten) * x[track_type]

        for track_type in x:
            x[track_type] = F.leaky_relu(
                self.selfloop_layers[-2][track_type](x[track_type])
            )

        for track_type in x:
            x[track_type] = F.leaky_relu(
                self.selfloop_layers[-1][track_type](x[track_type])
            )
        # print_heterogenous_representation(x)

        for idx, particle_type in enumerate(particle_types):
            if idx == 0:
                x_prime = x[particle_type]
                x_prime_batch = batch[particle_type].batch
            else:
                x_prime = torch.cat((x_prime, x[particle_type]), dim=0)
                x_prime_batch = torch.cat(
                    (x_prime_batch, batch[particle_type].batch), dim=0
                )

        x = global_mean_pool(x_prime, x_prime_batch.to(x_prime.device))

        if logits:
            return self.output_layer(x)
        else:
            disc_out = torch.sigmoid(self.output_layer(x))

        return disc_out

    # def forward(
    #     self,
    #     batch,  # contains all edge_index information
    # ) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
    #     # device = track_conditions.x.device

    #     particle_types = ["11", "13", "211", "321", "2212"]

    #     x = {}
    #     for track_type in particle_types:
    #         x[track_type] = torch.cat(
    #             (batch[f"{track_type}_conditions"].x, batch[f"{track_type}"].x), dim=1
    #         ).to(torch.float32)

    #         # print(batch[f"{track_type}_conditions"].x)
    #         # print(batch[f"{track_type}"].x)
    #         # print('a', x[track_type].dtype)

    #     edge_index_dict = {}
    #     edge_attr_dict = {}
    #     for edge_type in self.edge_types:
    #         edge_index_dict[edge_type[0], "to", edge_type[1]] = batch[
    #             edge_type[0], "to", edge_type[1]
    #         ]["edge_index"]
    #         edge_attr_dict[edge_type[0], "to", edge_type[1]] = batch[
    #             edge_type[0], "to", edge_type[1]
    #         ]["edge_attr"]

    #     # for i in x:
    #     #     x[i] = x[i].to('cuda:5')
    #     # for i in edge_index_dict:
    #     #     edge_index_dict[i] = edge_index_dict[i].to('cuda:5')
    #     # for i in edge_attr_dict:
    #     #     edge_attr_dict[i] = edge_attr_dict[i].to('cuda:5')

    #     x_comms = {}

    #     for N_layer in range(self.N_layers):
    #         for track_type in x:
    #             x[track_type] = F.leaky_relu(
    #                 self.selfloop_layers[N_layer][track_type](x[track_type])
    #             )
    #             x_comms[track_type] = x[track_type][:, : self.communication_dims]
    #         x_comms = self.communication_layers[N_layer](x_comms, edge_index_dict, edge_attr_dict)
    #         for track_type in x_comms:
    #             x_comms[track_type] = F.leaky_relu(x_comms[track_type])
    #         for track_type in x:
    #             x[track_type] = x[track_type].clone()
    #             x[track_type][:, : self.communication_dims] = x_comms[track_type]

    #     # for track_type in x:
    #     #     x[track_type] = F.leaky_relu(
    #     #         self.selfloop_layers[2][track_type](x[track_type])
    #     #     )
    #     #     x_comms[track_type] = x[track_type][:, : self.communication_dims]
    #     # x_comms = self.communication_layers[1](x_comms, edge_index_dict, edge_attr_dict)
    #     # for track_type in x_comms:
    #     #     x_comms[track_type] = F.leaky_relu(x_comms[track_type])
    #     # for track_type in x:
    #     #     x[track_type] = x[track_type].clone()
    #     #     x[track_type][:, : self.communication_dims] = x_comms[track_type]

    #     for track_type in x:
    #         x[track_type] = F.leaky_relu(
    #             self.selfloop_layers[-2][track_type](x[track_type])
    #         )
    #         if self.add_MBD:
    #             if x[track_type].shape[0] == 0:
    #                 mbd_output = torch.empty((0, self.mbd_dims)).to(x[track_type].device)
    #             else:
    #                 mbd_output = self.mbd_layer[track_type](x[track_type])
    #             x[track_type] = torch.cat((x[track_type], mbd_output), axis=1)

    #     for track_type in x:
    #         x[track_type] = F.leaky_relu(
    #             self.selfloop_layers[-1][track_type](x[track_type])
    #         )
    #     # print_heterogenous_representation(x)

    #     for idx, particle_type in enumerate(particle_types):
    #         if idx == 0:
    #             x_prime = x[particle_type]
    #             x_prime_batch = batch[particle_type].batch
    #         else:
    #             x_prime = torch.cat((x_prime, x[particle_type]), dim=0)
    #             x_prime_batch = torch.cat(
    #                 (x_prime_batch, batch[particle_type].batch), dim=0
    #             )

    #     x = global_mean_pool(x_prime, x_prime_batch.to(x_prime.device))

    #     # if self.add_MBD:
    #     #     mbd_output = self.mbd_layer(x)
    #     #     pooled_cat = torch.cat((x, mbd_output), axis=1)
    #     #     pooled_cat = x

    #     #     disc_out = torch.sigmoid(self.output_layer(pooled_cat))
    #     # else:
    #     disc_out = torch.sigmoid(self.output_layer(x))

    #     return disc_out


class gGAN(nn.Module):
    def __init__(
        self,
        track_targets_dims,
        track_conditions_dims,
        hidden_channels,
        track_latent_dims,
        edge_conditions_dims,
        silent=False,
        add_MBD=False,
        add_physics_dims=False,
    ):
        super(gGAN, self).__init__()

        if not silent:
            print("\n")
            print("Network params -----------------")
            print("track_targets_dims:", track_targets_dims)
            print("track_conditions_dims:", track_conditions_dims)
            print("hidden_channels:", hidden_channels)
            print("track_latent_dims:", track_latent_dims)
            print("edge_conditions_dims:", edge_conditions_dims)
            print("\n")

        self.track_latent_dims = track_latent_dims

        self.track_targets_dims = track_targets_dims

        self.generator = HeteroGenerator(
            track_targets_dims,
            track_conditions_dims,
            hidden_channels,
            track_latent_dims,
            edge_conditions_dims,
            conditonless=False,
        )

        self.discriminator = HeteroDiscriminator(
            track_targets_dims,
            track_conditions_dims,
            # int(hidden_channels/2.),
            hidden_channels,
            track_latent_dims,
            edge_conditions_dims,
            conditonless=False,
            add_MBD=add_MBD,
            add_physics_dims=add_physics_dims,
        )

    def inference(
        self, batch_size, batch, offline_query_mode=False, force_latent_value=None
    ):
        particle_types = ["11", "13", "211", "321", "2212"]
        track_latent = {}
        for particle_type in particle_types:
            if force_latent_value:
                track_latent[particle_type] = (
                    torch.ones(
                        (batch[particle_type].x.shape[0], self.track_latent_dims)
                    ).to(batch[particle_type].x.device)
                    * force_latent_value
                )
            else:
                track_latent[particle_type] = torch.randn(
                    (batch[particle_type].x.shape[0], self.track_latent_dims)
                ).to(batch[particle_type].x.device)

        out = self.generator(
            track_latent,
            batch=batch,
            offline_query_mode=offline_query_mode,
        )

        return out



import numpy as np
import vector

import uproot


def write_df_to_root(df, output_name, output_tree):
    df = df.loc[:, ~df.columns.duplicated()]

    branch_dict = {}
    data_dict = {}
    dtypes = df.dtypes
    used_columns = []  # Prevent repeat columns, kpipi_correction was getting repeated

    for branch in df.keys():
        if branch not in used_columns:
            dtype = dtypes[branch]

            # Convert unsigned types to signed
            if dtype == "uint32":
                dtype = "int32"
            elif dtype == "uint64":
                dtype = "int64"

            branch_dict[branch] = dtype

            # Handle DataFrame shapes
            if df[branch].ndim > 1:
                data_dict[branch] = (
                    df[branch].iloc[:, 0].to_numpy()
                )  # Convert to NumPy array
            else:
                data_dict[branch] = df[branch].to_numpy()  # Convert to NumPy array

            if isinstance(data_dict[branch][0], list):
                data_2d = np.vstack(data_dict[branch]).astype(dtype)

                # hack to somehow avoid uproot writing count_branch, n{branch_name}:
                vec = np.empty((np.shape(data_2d)[0], np.shape(data_2d)[1]))
                for dim in range(np.shape(data_2d)[1]):
                    vec[:, dim] = data_2d[:, dim]
                data_dict[branch] = vec

        used_columns.append(branch)

    # Create the ROOT file and write the tree
    with uproot.recreate(output_name) as f:
        # Create a new tree with the specified branches and types
        f[output_tree] = data_dict

    # if myGlobals._verbose:
    # print(f"Tuple written to {output_name}")


def compute_angles(origin, end):
    vector = np.array(end) - np.array(origin)
    theta = np.arccos(
        vector[2] / np.sqrt(vector[0] ** 2 + vector[1] ** 2 + vector[2] ** 2)
    )
    phi = np.arctan2(vector[1], vector[0])
    return theta, phi


def redefine_endpoint(origin, theta, phi, new_distance):
    x = new_distance * np.sin(theta) * np.cos(phi)
    y = new_distance * np.sin(theta) * np.sin(phi)
    z = new_distance * np.cos(theta)
    return origin[0] + x, origin[1] + y, origin[2] + z


def compute_DIRA(
    df, mother, particles, true_vars=True, true_vertex=False, RapidSim=False
):
    PX = 0
    PY = 0
    PZ = 0

    if true_vars:
        for particle in particles:
            if RapidSim:
                PX += df[f"{particle}_PX_TRUE"]
                PY += df[f"{particle}_PY_TRUE"]
                PZ += df[f"{particle}_PZ_TRUE"]
            else:
                PX += df[f"{particle}_TRUEP_X"]
                PY += df[f"{particle}_TRUEP_Y"]
                PZ += df[f"{particle}_TRUEP_Z"]

    else:
        for particle in particles:
            PX += df[f"{particle}_PX"]
            PY += df[f"{particle}_PY"]
            PZ += df[f"{particle}_PZ"]

    A = norm(np.asarray([PX, PY, PZ]))

    if RapidSim:
        if true_vertex:
            B = norm(
                np.asarray(
                    [
                        df[f"{mother}_vtxX_TRUE"] - df[f"{mother}_origX_TRUE"],
                        df[f"{mother}_vtxY_TRUE"] - df[f"{mother}_origY_TRUE"],
                        df[f"{mother}_vtxZ_TRUE"] - df[f"{mother}_origZ_TRUE"],
                    ]
                )
            )
        else:
            B = norm(
                np.asarray(
                    [
                        df[f"{mother}_vtxX"] - df[f"{mother}_origX"],
                        df[f"{mother}_vtxY"] - df[f"{mother}_origY"],
                        df[f"{mother}_vtxZ"] - df[f"{mother}_origZ"],
                    ]
                )
            )
    else:
        if true_vertex:
            B = norm(
                np.asarray(
                    [
                        df[f"{mother}_TRUEENDVERTEX_X"]
                        - df[f"{mother}_TRUEORIGINVERTEX_X"],
                        df[f"{mother}_TRUEENDVERTEX_Y"]
                        - df[f"{mother}_TRUEORIGINVERTEX_Y"],
                        df[f"{mother}_TRUEENDVERTEX_Z"]
                        - df[f"{mother}_TRUEORIGINVERTEX_Z"],
                    ]
                )
            )
        else:
            B = norm(
                np.asarray(
                    [
                        df[f"{mother}_ENDVERTEX_X"] - df[f"{mother}_OWNPV_X"],
                        df[f"{mother}_ENDVERTEX_Y"] - df[f"{mother}_OWNPV_Y"],
                        df[f"{mother}_ENDVERTEX_Z"] - df[f"{mother}_OWNPV_Z"],
                    ]
                )
            )

    dira = dot(A, B) / np.sqrt(mag(A) ** 2 * mag(B) ** 2)

    return dira


def compute_flightDistance(df, mother, particles, true_vars=True, true_vertex=False):
    # PX = 0
    # PY = 0
    # PZ = 0

    # if true_vars:
    #     for particle in particles:
    #         PX += df[f"{particle}_TRUEP_X"]
    #         PY += df[f"{particle}_TRUEP_Y"]
    #         PZ += df[f"{particle}_TRUEP_Z"]
    # else:
    #     for particle in particles:
    #         PX += df[f"{particle}_PX"]
    #         PY += df[f"{particle}_PY"]
    #         PZ += df[f"{particle}_PZ"]

    # momenta = vector.obj(
    #     px=PX,
    #     py=PY,
    #     pz=PZ,
    # )

    # if true_vertex:
    #     end_vertex = np.asarray(
    #         [df[f"{mother}_TRUEENDVERTEX_X"], df[f"{mother}_TRUEENDVERTEX_Y"], df[f"{mother}_TRUEENDVERTEX_Z"]]
    #     )
    #     primary_vertex = np.asarray(
    #         [df[f"{mother}_TRUEORIGINVERTEX_X"], df[f"{mother}_TRUEORIGINVERTEX_Y"], df[f"{mother}_TRUEORIGINVERTEX_Z"]]
    #     )
    # else:
    #     end_vertex = np.asarray(
    #         [df[f"{mother}_ENDVERTEX_X"], df[f"{mother}_ENDVERTEX_Y"], df[f"{mother}_ENDVERTEX_Z"]]
    #     )
    #     primary_vertex = np.asarray(
    #         [df[f"{mother}_OWNPV_X"], df[f"{mother}_OWNPV_Y"], df[f"{mother}_OWNPV_Z"]]
    #     )

    # momenta_array = np.asarray([momenta.px, momenta.py, momenta.pz])
    # P = momenta.mag
    # t = 0
    # for i in range(3):
    #     t += momenta_array[i] / P * (primary_vertex[i] - end_vertex[i])
    # dist = 0
    # for i in range(3):
    #     dist += (t * momenta_array[i] / P) ** 2
    # dist = np.sqrt(dist)

    dist = np.sqrt(
        (df[f"{mother}_TRUEENDVERTEX_X"] - df[f"{mother}_TRUEORIGINVERTEX_X"]) ** 2
        + (df[f"{mother}_TRUEENDVERTEX_Y"] - df[f"{mother}_TRUEORIGINVERTEX_Y"]) ** 2
        + (df[f"{mother}_TRUEENDVERTEX_Z"] - df[f"{mother}_TRUEORIGINVERTEX_Z"]) ** 2
    )
    return dist


def mag(vec):
    sum_sqs = 0
    for component in vec:
        sum_sqs += component**2
    mag = np.sqrt(sum_sqs)
    return mag


def norm(vec):
    mag_vec = mag(vec)
    for component_idx in range(np.shape(vec)[0]):
        vec[component_idx] *= 1.0 / mag_vec
    return vec


def dot(vec1, vec2):
    dot = 0
    for component_idx in range(np.shape(vec1)[0]):
        dot += vec1[component_idx] * vec2[component_idx]
    return dot


def compute_angle(df, mother, particle, true_vars=True, RapidSim=False):
    df = df.copy()
    if true_vars:
        if RapidSim:
            momenta_B = np.asarray(
                [
                    df[f"{mother}_PX_TRUE"],
                    df[f"{mother}_PY_TRUE"],
                    df[f"{mother}_PZ_TRUE"],
                ]
            )
            momenta_i = np.asarray(
                [
                    df[f"{particle}_PX_TRUE"],
                    df[f"{particle}_PY_TRUE"],
                    df[f"{particle}_PZ_TRUE"],
                ]
            )

        else:
            momenta_B = np.asarray(
                [
                    df[f"{mother}_TRUEP_X"],
                    df[f"{mother}_TRUEP_Y"],
                    df[f"{mother}_TRUEP_Z"],
                ]
            )
            momenta_i = np.asarray(
                [
                    df[f"{particle}_TRUEP_X"],
                    df[f"{particle}_TRUEP_Y"],
                    df[f"{particle}_TRUEP_Z"],
                ]
            )
    else:
        momenta_B = np.asarray(
            [df[f"{mother}_PX"], df[f"{mother}_PY"], df[f"{mother}_PZ"]]
        )
        momenta_i = np.asarray(
            [
                df[f"{particle}_PX"],
                df[f"{particle}_PY"],
                df[f"{particle}_PZ"],
            ]
        )

    dot_prod = np.arccos(dot(momenta_B, momenta_i) / (mag(momenta_B) * mag(momenta_i)))

    dot_prod[np.where(np.isnan(dot_prod))] = 1e-6
    dot_prod[np.where(dot_prod == 0)] = 1e-6

    return dot_prod


def compute_impactParameter_i(
    df, mother, particle, true_vars=True, true_vertex=False, RapidSim=False
):
    df = df.copy()

    if true_vars:
        if RapidSim:
            momenta = vector.obj(
                px=df[f"{particle}_PX_TRUE"],
                py=df[f"{particle}_PY_TRUE"],
                pz=df[f"{particle}_PZ_TRUE"],
            )
        else:
            momenta = vector.obj(
                px=df[f"{particle}_TRUEP_X"],
                py=df[f"{particle}_TRUEP_Y"],
                pz=df[f"{particle}_TRUEP_Z"],
            )

    else:
        momenta = vector.obj(
            px=df[f"{particle}_PX"],
            py=df[f"{particle}_PY"],
            pz=df[f"{particle}_PZ"],
        )

    if RapidSim:
        if true_vertex:
            end_vertex = np.asarray(
                [
                    df[f"{mother}_vtxX_TRUE"],
                    df[f"{mother}_vtxY_TRUE"],
                    df[f"{mother}_vtxZ_TRUE"],
                ]
            )
            primary_vertex = np.asarray(
                [
                    df[f"{mother}_origX_TRUE"],
                    df[f"{mother}_origY_TRUE"],
                    df[f"{mother}_origZ_TRUE"],
                ]
            )
        else:
            end_vertex = np.asarray(
                [
                    df[f"{mother}_vtxX"],
                    df[f"{mother}_vtxY"],
                    df[f"{mother}_vtxZ"],
                ]
            )
            primary_vertex = np.asarray(
                [df[f"{mother}_origX"], df[f"{mother}_origY"], df[f"{mother}_origZ"]]
            )
    else:
        if true_vertex:
            end_vertex = np.asarray(
                [
                    df[f"{mother}_TRUEENDVERTEX_X"],
                    df[f"{mother}_TRUEENDVERTEX_Y"],
                    df[f"{mother}_TRUEENDVERTEX_Z"],
                ]
            )
            primary_vertex = np.asarray(
                [
                    df[f"{mother}_TRUEORIGINVERTEX_X"],
                    df[f"{mother}_TRUEORIGINVERTEX_Y"],
                    df[f"{mother}_TRUEORIGINVERTEX_Z"],
                ]
            )
        else:
            end_vertex = np.asarray(
                [
                    df[f"{mother}_ENDVERTEX_X"],
                    df[f"{mother}_ENDVERTEX_Y"],
                    df[f"{mother}_ENDVERTEX_Z"],
                ]
            )
            primary_vertex = np.asarray(
                [
                    df[f"{mother}_OWNPV_X"],
                    df[f"{mother}_OWNPV_Y"],
                    df[f"{mother}_OWNPV_Z"],
                ]
            )

    momenta_array = np.asarray([momenta.px, momenta.py, momenta.pz])
    P = momenta.mag
    t = 0
    for i in range(3):
        t += momenta_array[i] / P * (primary_vertex[i] - end_vertex[i])
    dist = 0
    for i in range(3):
        dist += (primary_vertex[i] - end_vertex[i] - t * momenta_array[i] / P) ** 2
    dist = np.sqrt(dist)

    return dist


def compute_intermediate_distance(df, intermediate, mother):
    df = df.copy()
    X = df[f"{intermediate}_TRUEENDVERTEX_X"] - df[f"{mother}_TRUEENDVERTEX_X"]
    Y = df[f"{intermediate}_TRUEENDVERTEX_Y"] - df[f"{mother}_TRUEENDVERTEX_Y"]
    Z = df[f"{intermediate}_TRUEENDVERTEX_Z"] - df[f"{mother}_TRUEENDVERTEX_Z"]
    dist = np.sqrt(X**2 + Y**2 + Z**2)
    return dist


def compute_distance(df, A, A_var, B, B_var, RapidSim=False, true_vars=True):
    df = df.copy()

    if RapidSim:
        X = df[f"{A}_{A_var}X_TRUE"] - df[f"{B}_{B_var}X_TRUE"]
        Y = df[f"{A}_{A_var}Y_TRUE"] - df[f"{B}_{B_var}Y_TRUE"]
        Z = df[f"{A}_{A_var}Z_TRUE"] - df[f"{B}_{B_var}Z_TRUE"]
    else:
        X = df[f"{A}_{A_var}_X"] - df[f"{B}_{B_var}_X"]
        Y = df[f"{A}_{A_var}_Y"] - df[f"{B}_{B_var}_Y"]
        Z = df[f"{A}_{A_var}_Z"] - df[f"{B}_{B_var}_Z"]

    dist = np.sqrt(X**2 + Y**2 + Z**2)
    return dist


def compute_impactParameter(
    df, mother, particles, true_vars=True, true_vertex=False, RapidSim=False
):
    df = df.copy()

    PX = 0
    PY = 0
    PZ = 0

    if true_vars:
        for particle in particles:
            if RapidSim:
                PX += df[f"{particle}_PX_TRUE"]
                PY += df[f"{particle}_PY_TRUE"]
                PZ += df[f"{particle}_PZ_TRUE"]
            else:
                PX += df[f"{particle}_TRUEP_X"]
                PY += df[f"{particle}_TRUEP_Y"]
                PZ += df[f"{particle}_TRUEP_Z"]
    else:
        for particle in particles:
            PX += df[f"{particle}_PX"]
            PY += df[f"{particle}_PY"]
            PZ += df[f"{particle}_PZ"]

    momenta = vector.obj(
        px=PX,
        py=PY,
        pz=PZ,
    )
    if RapidSim:
        if true_vertex:
            end_vertex = np.asarray(
                [
                    df[f"{mother}_vtxX_TRUE"],
                    df[f"{mother}_vtxY_TRUE"],
                    df[f"{mother}_vtxZ_TRUE"],
                ]
            )
            primary_vertex = np.asarray(
                [
                    df[f"{mother}_origX_TRUE"],
                    df[f"{mother}_origY_TRUE"],
                    df[f"{mother}_origZ_TRUE"],
                ]
            )
        else:
            end_vertex = np.asarray(
                [
                    df[f"{mother}_vtxX"],
                    df[f"{mother}_vtxY"],
                    df[f"{mother}_vtxZ"],
                ]
            )
            primary_vertex = np.asarray(
                [df[f"{mother}_origX"], df[f"{mother}_origY"], df[f"{mother}_origZ"]]
            )
    else:
        if true_vertex:
            end_vertex = np.asarray(
                [
                    df[f"{mother}_TRUEENDVERTEX_X"],
                    df[f"{mother}_TRUEENDVERTEX_Y"],
                    df[f"{mother}_TRUEENDVERTEX_Z"],
                ]
            )
            primary_vertex = np.asarray(
                [
                    df[f"{mother}_TRUEORIGINVERTEX_X"],
                    df[f"{mother}_TRUEORIGINVERTEX_Y"],
                    df[f"{mother}_TRUEORIGINVERTEX_Z"],
                ]
            )
        else:
            end_vertex = np.asarray(
                [
                    df[f"{mother}_ENDVERTEX_X"],
                    df[f"{mother}_ENDVERTEX_Y"],
                    df[f"{mother}_ENDVERTEX_Z"],
                ]
            )
            primary_vertex = np.asarray(
                [
                    df[f"{mother}_OWNPV_X"],
                    df[f"{mother}_OWNPV_Y"],
                    df[f"{mother}_OWNPV_Z"],
                ]
            )

    momenta_array = np.asarray([momenta.px, momenta.py, momenta.pz])
    P = momenta.mag
    t = 0
    for i in range(3):
        t += momenta_array[i] / P * (primary_vertex[i] - end_vertex[i])
    dist = 0
    for i in range(3):
        dist += (primary_vertex[i] - end_vertex[i] - t * momenta_array[i] / P) ** 2
    dist = np.sqrt(dist)

    return dist


# def compute_delta_mom(df, particle):
def compute_reconstructed_momentum_residual(df, particle, RapidSim=False):
    df = df.copy()

    if RapidSim:
        PX = df[f"{particle}_PX"] - df[f"{particle}_PX_TRUE"]
        PY = df[f"{particle}_PY"] - df[f"{particle}_PY_TRUE"]
        PZ = df[f"{particle}_PZ"] - df[f"{particle}_PZ_TRUE"]
    else:
        PX = df[f"{particle}_PX"] - df[f"{particle}_TRUEP_X"]
        PY = df[f"{particle}_PY"] - df[f"{particle}_TRUEP_Y"]
        PZ = df[f"{particle}_PZ"] - df[f"{particle}_TRUEP_Z"]

    return np.sqrt((PX**2 + PY**2 + PZ**2)), np.sqrt((PX**2 + PY**2))


# def compute_miss_mom(df, mother, particles, true_vars=True):
def compute_missing_momentum(df, mother, particles, true_vars=True, RapidSim=False):
    df = df.copy()

    if true_vars:
        if RapidSim:
            PX = np.asarray(df[f"{mother}_PX_TRUE"]).copy()
            PY = np.asarray(df[f"{mother}_PY_TRUE"]).copy()
            PZ = np.asarray(df[f"{mother}_PZ_TRUE"]).copy()

            for particle in particles:
                PX += -1.0 * np.asarray(df[f"{particle}_PX_TRUE"]).copy()
                PY += -1.0 * np.asarray(df[f"{particle}_PY_TRUE"]).copy()
                PZ += -1.0 * np.asarray(df[f"{particle}_PZ_TRUE"]).copy()
        else:
            PX = np.asarray(df[f"{mother}_TRUEP_X"]).copy()
            PY = np.asarray(df[f"{mother}_TRUEP_Y"]).copy()
            PZ = np.asarray(df[f"{mother}_TRUEP_Z"]).copy()

            for particle in particles:
                PX += -1.0 * np.asarray(df[f"{particle}_TRUEP_X"]).copy()
                PY += -1.0 * np.asarray(df[f"{particle}_TRUEP_Y"]).copy()
                PZ += -1.0 * np.asarray(df[f"{particle}_TRUEP_Z"]).copy()
    else:
        PX = np.asarray(df[f"{mother}_PX"]).copy()
        PY = np.asarray(df[f"{mother}_PY"]).copy()
        PZ = np.asarray(df[f"{mother}_PZ"]).copy()

        for particle in particles:
            PX += -1.0 * np.asarray(df[f"{particle}_PX"]).copy()
            PY += -1.0 * np.asarray(df[f"{particle}_PY"]).copy()
            PZ += -1.0 * np.asarray(df[f"{particle}_PZ"]).copy()

    return np.sqrt((PX**2 + PY**2 + PZ**2)), np.sqrt((PX**2 + PY**2))


def compute_reconstructed_intermediate_momenta(df, particles, true_vars=True):
    df = df.copy()

    PX = 0
    PY = 0
    PZ = 0

    if true_vars:
        for particle in particles:
            PX += df[f"{particle}_TRUEP_X"]
            PY += df[f"{particle}_TRUEP_Y"]
            PZ += df[f"{particle}_TRUEP_Z"]
    else:
        for particle in particles:
            PX += df[f"{particle}_PX"]
            PY += df[f"{particle}_PY"]
            PZ += df[f"{particle}_PZ"]

    return np.sqrt((PX**2 + PY**2 + PZ**2)), np.sqrt((PX**2 + PY**2))


def compute_TRUEE(df, particle, RapidSim=False):
    df = df.copy()

    M = df[f"{particle}_mass"]
    if RapidSim:
        PX = df[f"{particle}_PX_TRUE"]
        PY = df[f"{particle}_PY_TRUE"]
        PZ = df[f"{particle}_PZ_TRUE"]
    else:
        PX = df[f"{particle}_TRUEP_X"]
        PY = df[f"{particle}_TRUEP_Y"]
        PZ = df[f"{particle}_TRUEP_Z"]

    return np.sqrt((M**2 + PX**2 + PY**2 + PZ**2))


def compute_TRUEP(df, particle, RapidSim=False):
    df = df.copy()

    if RapidSim:
        PX = df[f"{particle}_PX_TRUE"]
        PY = df[f"{particle}_PY_TRUE"]
        PZ = df[f"{particle}_PZ_TRUE"]
    else:
        PX = df[f"{particle}_TRUEP_X"]
        PY = df[f"{particle}_TRUEP_Y"]
        PZ = df[f"{particle}_TRUEP_Z"]

    return np.sqrt((PX**2 + PY**2 + PZ**2))


# def compute_B_mom(df, mother, true_vars=True):
def compute_reconstructed_mother_momenta(df, mother, true_vars=True):
    df = df.copy()

    if true_vars:
        PX = df[f"{mother}_TRUEP_X"]
        PY = df[f"{mother}_TRUEP_Y"]
        PZ = df[f"{mother}_TRUEP_Z"]
    else:
        PX = df[f"{mother}_PX"]
        PY = df[f"{mother}_PY"]
        PZ = df[f"{mother}_PZ"]

    return np.sqrt((PX**2 + PY**2 + PZ**2)), np.sqrt((PX**2 + PY**2))


def compute_mass_N(df, particles, true_vars=True):
    df = df.copy()

    if true_vars:
        for p_idx, particle in enumerate(particles):
            if p_idx == 0:
                PE = np.sqrt(
                    df[f"{particle}_mass"] ** 2
                    + df[f"{particle}_TRUEP_X"] ** 2
                    + df[f"{particle}_TRUEP_Y"] ** 2
                    + df[f"{particle}_TRUEP_Z"] ** 2
                )
                PX = df[f"{particle}_TRUEP_X"]
                PY = df[f"{particle}_TRUEP_Y"]
                PZ = df[f"{particle}_TRUEP_Z"]

            else:
                PE += np.sqrt(
                    df[f"{particle}_mass"] ** 2
                    + df[f"{particle}_TRUEP_X"] ** 2
                    + df[f"{particle}_TRUEP_Y"] ** 2
                    + df[f"{particle}_TRUEP_Z"] ** 2
                )
                PX += df[f"{particle}_TRUEP_X"]
                PY += df[f"{particle}_TRUEP_Y"]
                PZ += df[f"{particle}_TRUEP_Z"]
    else:
        for p_idx, particle in enumerate(particles):
            if p_idx == 0:
                PE = np.sqrt(
                    df[f"{particle}_mass"] ** 2
                    + df[f"{particle}_PX"] ** 2
                    + df[f"{particle}_PY"] ** 2
                    + df[f"{particle}_PZ"] ** 2
                )
                PX = df[f"{particle}_PX"]
                PY = df[f"{particle}_PY"]
                PZ = df[f"{particle}_PZ"]

            else:
                PE += np.sqrt(
                    df[f"{particle}_mass"] ** 2
                    + df[f"{particle}_PX"] ** 2
                    + df[f"{particle}_PY"] ** 2
                    + df[f"{particle}_PZ"] ** 2
                )
                PX += df[f"{particle}_PX"]
                PY += df[f"{particle}_PY"]
                PZ += df[f"{particle}_PZ"]

    mass = np.sqrt((PE**2 - PX**2 - PY**2 - PZ**2))

    return mass



from rich.console import Console
from lhcb_rex.tools.stopwatch import stopwatch
import os
import pandas as pd
import torch

particle_universe = [11, 13, 211, 321, 2212]

masses = {}
masses[321] = 493.677
masses[211] = 139.57039
masses[13] = 105.66
masses[11] = 0.51099895000
masses[2212] = 938.272

particle_masses_dict_mothers = {}
particle_masses_dict_mothers[411] = 1869.62
particle_masses_dict_mothers[421] = 1864.84
particle_masses_dict_mothers[431] = 1968.47

particle_masses_dict_mothers[511] = 5279.65
particle_masses_dict_mothers[521] = 5279.34
particle_masses_dict_mothers[531] = 5366.88
particle_masses_dict_mothers[541] = 6274.9


masses_GeV = {}
masses_GeV[321] = 493.677 * 1e-3
masses_GeV[211] = 139.57039 * 1e-3
masses_GeV[13] = 105.66 * 1e-3
masses_GeV[11] = 0.51099895000 * 1e-3
masses_GeV[2212] = 938.272 * 1e-3

particle_masses_dict_mothers_GeV = {}
particle_masses_dict_mothers_GeV[411] = 1869.62 * 1e-3
particle_masses_dict_mothers_GeV[421] = 1864.84 * 1e-3
particle_masses_dict_mothers_GeV[431] = 1968.47 * 1e-3

particle_masses_dict_mothers_GeV[511] = 5279.65 * 1e-3
particle_masses_dict_mothers_GeV[521] = 5279.34 * 1e-3
particle_masses_dict_mothers_GeV[531] = 5366.88 * 1e-3
particle_masses_dict_mothers_GeV[541] = 6274.9 * 1e-3


personalised_track_node_types = False

if torch.cuda.is_available():
    device = torch.device("cuda:5")
else:
    device = torch.device("cpu")
    print("EXECUTING ON CPU")


stopwatches = stopwatch()

console = Console()

my_decay_splash = []


# Check environment variables
def get_environment_variable(var_name: str) -> str:
    value = os.environ.get(var_name)
    if value is None:
        raise EnvironmentError(f"Environment variable ${var_name} not set.")
    return value


rapidsim_settings = {}
rapidsim_settings["imported"] = False

if os.environ.get("RAPIDSIM_ROOT"):
    PARTICLES_FILE = "config/particles.dat"
    EVTGEN_ROOT_ENV = "EVTGEN_ROOT"
    RAPIDSIM_EXECUTABLE = "/build/src/RapidSim.exe"

    rapidsim_settings["RAPIDSIM_ROOT"] = os.environ.get("RAPIDSIM_ROOT")
    rapidsim_settings["RAPIDSIM_particles"] = (
        os.environ.get("RAPIDSIM_ROOT") + "/" + PARTICLES_FILE
    )
    rapidsim_settings["RAPIDSIM_particles_df"] = pd.read_csv(
        rapidsim_settings["RAPIDSIM_particles"], sep=r"\s+"
    )
    rapidsim_settings["USE_EVTGEN"] = (
        get_environment_variable(EVTGEN_ROOT_ENV) is not None
    )
    rapidsim_settings["rapidsim_exe"] = (
        f"{rapidsim_settings['RAPIDSIM_ROOT']}{RAPIDSIM_EXECUTABLE}"
    )
    rapidsim_settings["imported"] = True

else:
    print("RapidSim not present")
    # Running without RapidSim


def global_print_my_decay_splash():
    if len(my_decay_splash) > 0:
        for counts, line in enumerate(my_decay_splash):
            if counts % 2 == 0:
                console.print("")
            console.print(line)
        console.print("")


smearPV_targets = [
    "MOTHER_TRUE_FD",
    # "MOTHER_TRUEENDVERTEX_X",
    # "MOTHER_TRUEENDVERTEX_Y",
    # "MOTHER_TRUEENDVERTEX_Z",
    "MOTHER_TRUEORIGINVERTEX_X",
    "MOTHER_TRUEORIGINVERTEX_Y",
    "MOTHER_TRUEORIGINVERTEX_Z",
]
smearPV_conditions = [
    "MOTHER_TRUEP",
    "MOTHER_TRUEP_T",
    "MOTHER_TRUEP_X",
    "MOTHER_TRUEP_Y",
    "MOTHER_TRUEP_Z",
]


# smearPV_targets = [
#     "MOTHER_TRUEORIGINVERTEX_X",
#     "MOTHER_TRUEORIGINVERTEX_Y",
#     "MOTHER_TRUEORIGINVERTEX_Z",
#     "MOTHER_OWNPV_X",
#     "MOTHER_OWNPV_Y",
#     "MOTHER_OWNPV_Z",
#     "MOTHER_TRUEENDVERTEX_X",
#     "MOTHER_TRUEENDVERTEX_Y",
#     "MOTHER_TRUEENDVERTEX_Z",
#     "MOTHER_ENDVERTEX_X",
#     "MOTHER_ENDVERTEX_Y",
#     "MOTHER_ENDVERTEX_Z",
# ]
# smearPV_conditions = [
#     "MOTHER_FLIGHT",
# ]


smearing_track_targets = [
    "DAUGHTERN_delta_PX",
    "DAUGHTERN_delta_PY",
    # "DAUGHTERN_Preco_overP",
    "DAUGHTERN_delta_PZ",
    "DAUGHTERN_TRACK_CHI2NDOF",
    "DAUGHTERN_TRACK_GhostProb",
]
smearing_track_conditions = [
    "DAUGHTERN_TRUEP_X",
    "DAUGHTERN_TRUEP_Y",
    "DAUGHTERN_TRUEP_Z",
    "DAUGHTERN_TRUEP",
    "DAUGHTERN_TRUEE",
]
smearing_edge_conditions = [
    "edge_angle_DAUGHTERN_DAUGHTERN_TRUE",
]


PID_track_targets = [
    "DAUGHTERN_PIDe",
    "DAUGHTERN_PIDmu",
    "DAUGHTERN_PIDK",
    "DAUGHTERN_PIDp",
    "DAUGHTERN_MC15TuneV1_ProbNNe",
    "DAUGHTERN_MC15TuneV1_ProbNNmu",
    "DAUGHTERN_MC15TuneV1_ProbNNpi",
    "DAUGHTERN_MC15TuneV1_ProbNNk",
    "DAUGHTERN_MC15TuneV1_ProbNNp",
    # "DAUGHTERN_L0HadronDecision_TIS",
    # "DAUGHTERN_L0HadronDecision_TOS",
    # "DAUGHTERN_L0ElectronDecision_TIS",
    # "DAUGHTERN_L0ElectronDecision_TOS",
    # "DAUGHTERN_L0MuonDecision_TIS",
    # "DAUGHTERN_L0MuonDecision_TOS",
    # "DAUGHTERN_L0PhotonDecision_TIS",
    # "DAUGHTERN_L0PhotonDecision_TOS",
]

PID_track_conditions = [
    "DAUGHTERN_TRUEP_X",
    "DAUGHTERN_TRUEP_Y",
    "DAUGHTERN_TRUEP_Z",
    "DAUGHTERN_TRUEP",
    "DAUGHTERN_TRUEE",
    "DAUGHTERN_delta_PX",
    "DAUGHTERN_delta_PY",
    "DAUGHTERN_Preco_overP",
    "DAUGHTERN_IP",
    "DAUGHTERN_IP_TRUE",
    "DAUGHTERN_angle_wrt_mother",
    "DAUGHTERN_angle_wrt_mother_reco",
    "DAUGHTERN_eta",
    "DAUGHTERN_eta_TRUE",
    "DAUGHTERN_TRACK_CHI2NDOF",
    "DAUGHTERN_TRACK_GhostProb",
]
PID_edge_conditions = [
    "edge_angle_DAUGHTERN_DAUGHTERN_TRUE",
]


option = "A"  # standard
# option = "B" # no vertex information
# option = "C" # predict vertex information


mother_targets = [
    "MOTHER_VTXISOBDTHARDFIRSTVALUE",
    "MOTHER_VTXISOBDTHARDSECONDVALUE",
    "MOTHER_VTXISOBDTHARDTHIRDVALUE",
    "MOTHER_ENDVERTEX_CHI2NDOF",
    # "MOTHER_ENDVERTEX_CHI2",
    "MOTHER_IPCHI2_OWNPV",
    "MOTHER_FDCHI2_OWNPV",
    "MOTHER_DIRA_OWNPV",
]
intermediate_targets = [
    "INTERMEDIATE_ENDVERTEX_CHI2NDOF",
    # "INTERMEDIATE_ENDVERTEX_CHI2",
    "INTERMEDIATE_IPCHI2_OWNPV",
    "INTERMEDIATE_FDCHI2_OWNPV",
    "INTERMEDIATE_DIRA_OWNPV",
]
track_targets = [
    "DAUGHTERN_IPCHI2_OWNPV",
    # "DAUGHTERN_TRACK_CHI2NDOF",
    # "DAUGHTERN_TRACK_GhostProb",
]

if option == "C":
    mother_targets.extend(
        [
            "MOTHER_TRUEENDVERTEX_X",
            "MOTHER_TRUEENDVERTEX_Y",
            "MOTHER_TRUEENDVERTEX_Z",
            "MOTHER_TRUEORIGINVERTEX_X",
            "MOTHER_TRUEORIGINVERTEX_Y",
            "MOTHER_TRUEORIGINVERTEX_Z",
            "MOTHER_DIRA",
            "MOTHER_DIRA_TRUE",
            "MOTHER_IP",
            "MOTHER_IP_TRUE",
        ]
    )
    intermediate_targets.extend(
        [
            "INTERMEDIATE_DIRA",
            "INTERMEDIATE_DIRA_TRUE",
            "INTERMEDIATE_IP",
            "INTERMEDIATE_IP_TRUE",
        ]
    )
    track_targets.extend(
        [
            "DAUGHTERN_IP",
            "DAUGHTERN_IP_TRUE",
        ]
    )

if option == "A":
    mother_conditions = [  # these things are not accessible to intermediate - intermediate might not exist
        "MOTHER_TRUEENDVERTEX_X",
        "MOTHER_TRUEENDVERTEX_Y",
        "MOTHER_TRUEENDVERTEX_Z",
        # "MOTHER_TRUEORIGINVERTEX_X",
        # "MOTHER_TRUEORIGINVERTEX_Y",
        "MOTHER_TRUEORIGINVERTEX_Z",
        "MOTHER_TRUEP",
        "MOTHER_TRUEP_T",
        "MOTHER_TRUEP_X",
        "MOTHER_TRUEP_Y",
        "MOTHER_TRUEP_Z",
        "MOTHER_FLIGHT",
        "MOTHER_DIRA",
        "MOTHER_DIRA_TRUE",
        "MOTHER_IP",
        "MOTHER_IP_TRUE",
        "MOTHER_P",
        "MOTHER_PT",
        "MOTHER_PX",
        "MOTHER_PY",
        "MOTHER_PZ",
        "MOTHER_missing_P", # in new arch
        "MOTHER_missing_PT", # in new arch
        # "N_daughters",
    ]  # N_daughters must be the last condition
    intermediate_conditions = [
        "INTERMEDIATE_DIRA",
        "INTERMEDIATE_DIRA_TRUE",
        "INTERMEDIATE_IP",
        "INTERMEDIATE_IP_TRUE",
        "INTERMEDIATE_P",
        "INTERMEDIATE_PT",
        "INTERMEDIATE_PX",
        "INTERMEDIATE_PY",
        "INTERMEDIATE_PZ",
        # "N_daughters",
    ]  # N_daughters must be the last condition
    track_conditions = [
        "DAUGHTERN_FLIGHT",
        "DAUGHTERN_IP",
        "DAUGHTERN_IP_TRUE",
        "DAUGHTERN_PX",
        "DAUGHTERN_PY",
        "DAUGHTERN_PZ",
        "DAUGHTERN_TRUEP_X",
        "DAUGHTERN_TRUEP_Y",
        "DAUGHTERN_TRUEP_Z",
        "DAUGHTERN_angle_wrt_mother",
        "DAUGHTERN_angle_wrt_mother_reco",
        "DAUGHTERN_delta_P", # in new arch
        "DAUGHTERN_delta_PT", # in new arch
        "DAUGHTERN_eta",
        "DAUGHTERN_eta_TRUE",
        # "DAUGHTERN_residualfrac_PX",
        # "DAUGHTERN_residualfrac_PY",
        # "DAUGHTERN_residualfrac_PZ",
        # # PID variables
        # "DAUGHTERN_PIDe",
        # "DAUGHTERN_PIDmu",
        # "DAUGHTERN_PIDK",
        # "DAUGHTERN_PIDp",
        # "DAUGHTERN_MC15TuneV1_ProbNNe",
        # "DAUGHTERN_MC15TuneV1_ProbNNmu",
        # "DAUGHTERN_MC15TuneV1_ProbNNpi",
        # "DAUGHTERN_MC15TuneV1_ProbNNk",
        # "DAUGHTERN_MC15TuneV1_ProbNNp",
        "DAUGHTERN_TRACK_CHI2NDOF", # in new arch
        "DAUGHTERN_TRACK_GhostProb", # in new arch
    ]
else:
    mother_conditions = [  # these things are not accessible to intermediate - intermediate might not exist
        # "MOTHER_TRUEENDVERTEX_X",
        # "MOTHER_TRUEENDVERTEX_Y",
        # "MOTHER_TRUEENDVERTEX_Z",
        # "MOTHER_TRUEORIGINVERTEX_X",
        # "MOTHER_TRUEORIGINVERTEX_Y",
        # "MOTHER_TRUEORIGINVERTEX_Z",
        "MOTHER_TRUEP",
        "MOTHER_TRUEP_T",
        "MOTHER_TRUEP_X",
        "MOTHER_TRUEP_Y",
        "MOTHER_TRUEP_Z",
        "MOTHER_FLIGHT",
        # "MOTHER_DIRA",
        # "MOTHER_DIRA_TRUE",
        # "MOTHER_IP",
        # "MOTHER_IP_TRUE",
        "MOTHER_P",
        "MOTHER_PT",
        "MOTHER_PX",
        "MOTHER_PY",
        "MOTHER_PZ",
        "MOTHER_missing_P",
        "MOTHER_missing_PT",
        "N_daughters",
    ]  # N_daughters must be the last condition
    intermediate_conditions = [
        # "INTERMEDIATE_DIRA",
        # "INTERMEDIATE_DIRA_TRUE",
        # "INTERMEDIATE_IP",
        # "INTERMEDIATE_IP_TRUE",
        "INTERMEDIATE_P",
        "INTERMEDIATE_PT",
        "INTERMEDIATE_PX",
        "INTERMEDIATE_PY",
        "INTERMEDIATE_PZ",
        "N_daughters",
    ]  # N_daughters must be the last condition
    track_conditions = [
        "DAUGHTERN_FLIGHT",
        # "DAUGHTERN_IP",
        # "DAUGHTERN_IP_TRUE",
        "DAUGHTERN_PX",
        "DAUGHTERN_PY",
        "DAUGHTERN_PZ",
        "DAUGHTERN_TRUEP_X",
        "DAUGHTERN_TRUEP_Y",
        "DAUGHTERN_TRUEP_Z",
        "DAUGHTERN_angle_wrt_mother",
        "DAUGHTERN_angle_wrt_mother_reco",
        "DAUGHTERN_delta_P",
        "DAUGHTERN_delta_PT",
        "DAUGHTERN_eta",
        "DAUGHTERN_eta_TRUE",
        "DAUGHTERN_residualfrac_PX",
        "DAUGHTERN_residualfrac_PY",
        "DAUGHTERN_residualfrac_PZ",
        # # PID variables
        # "DAUGHTERN_PIDe",
        # "DAUGHTERN_PIDmu",
        # "DAUGHTERN_PIDK",
        # "DAUGHTERN_PIDp",
        # "DAUGHTERN_MC15TuneV1_ProbNNe",
        # "DAUGHTERN_MC15TuneV1_ProbNNmu",
        # "DAUGHTERN_MC15TuneV1_ProbNNpi",
        # "DAUGHTERN_MC15TuneV1_ProbNNk",
        # "DAUGHTERN_MC15TuneV1_ProbNNp",
        # "DAUGHTERN_TRACK_CHI2NDOF",
        # "DAUGHTERN_TRACK_GhostProb",
    ]

validation_variables = [
    "mkl",
    "q2",
    "mkee",
    "ctl",
]



import lhcb_rex.settings.globals as myGlobals
from termcolor import colored
from pathlib import Path
from contextlib import contextmanager
from datetime import datetime
from rich.table import Table
import uproot
import os
import string
import re


def print_decay_splash(
    decay_strings,
    daughter_string,
    label_string,
    ordering_string,
    naming_scheme,
    mass_hypotheses,
    model_scheme,
):
    # myGlobals.console.print("")

    # Create hierarchy_string and hierarchy_labels using letters
    hierarchy_string = f"{decay_strings[0]} -> {daughter_string}"
    hierarchy_labels = f"{len(decay_strings[0]) * 'a'} cc "

    # Convert numerical labels to letters
    for character in label_string:
        if character == " ":
            hierarchy_labels += character
        else:
            hierarchy_labels += string.ascii_lowercase[(int(character) + 1) * 2]

    bracketless_hierarchy_string = ""
    bracketless_hierarchy_labels = ""

    # Iterate through the characters of hierarchy_string and hierarchy_labels together
    for char, label in zip(hierarchy_string, hierarchy_labels):
        if char not in ["{", "}"]:
            bracketless_hierarchy_string += char
            bracketless_hierarchy_labels += label

    # Remove double spaces
    bracketless_hierarchy_string = bracketless_hierarchy_string.replace("  ", " ")
    bracketless_hierarchy_labels = bracketless_hierarchy_labels.replace("  ", " ")

    bracketless_ordering_string = ordering_string.replace("{", "")
    bracketless_ordering_string = bracketless_ordering_string.replace("}", "")
    bracketless_ordering_string = bracketless_ordering_string.replace("  ", " ")

    labelled_bracketless_hierarchy_string = ""
    labelled_bracketless_hierarchy_labels = ""
    labelled_dim_string = ""
    for particle, particle_labels, ordering_i in zip(
        bracketless_hierarchy_string.split(" "),
        bracketless_hierarchy_labels.split(" "),
        bracketless_ordering_string.split(" "),
    ):
        if particle != " ":
            if particle == "->":
                labelled_bracketless_hierarchy_string += f" {particle}"
                labelled_bracketless_hierarchy_labels += f" {particle_labels}"
                labelled_dim_string += f" {len(particle_labels) * 'd'}"
            else:
                if len(particle) > 0:
                    current_label = particle_labels[0]

                    labelled_bracketless_hierarchy_string += f" {particle}"
                    labelled_bracketless_hierarchy_labels += f" {particle_labels}"

                    # now add label
                    next_label = (
                        string.ascii_lowercase[
                            string.ascii_lowercase.index(current_label) + 1
                        ]
                        if current_label != "z"
                        else "a"
                    )

                    label_i = f"{naming_scheme[int(ordering_i)]}"

                    if model_scheme[int(ordering_i)] != "PHSP":
                        label_i += f", {model_scheme[int(ordering_i)].split(' ')[0]}"
                    if mass_hypotheses is not None:
                        if naming_scheme[int(ordering_i)] in mass_hypotheses:
                            label_i += (
                                f", {mass_hypotheses[naming_scheme[int(ordering_i)]]}"
                            )

                    to_append = f"[{label_i}]  "
                    # if naming_scheme[int(ordering_i)] != 'NA': # this particle is reconstructed
                    if (
                        not re.match(r"^NA_\d{8}$", naming_scheme[int(ordering_i)])
                        and naming_scheme[int(ordering_i)] != "NA"
                    ):  # this particle is reconstructed
                        labelled_bracketless_hierarchy_string += to_append
                        labelled_bracketless_hierarchy_labels += (
                            f"{len(to_append) * next_label}"
                        )
                        labelled_dim_string += (
                            f" {len(particle_labels) * 'n'}{len(to_append) * 'n'}"
                        )
                    else:
                        labelled_dim_string += f" {len(particle_labels) * 'd'}"

    labelled_bracketless_hierarchy_string = labelled_bracketless_hierarchy_string[1:]
    labelled_bracketless_hierarchy_labels = labelled_bracketless_hierarchy_labels[1:]
    labelled_dim_string = labelled_dim_string[1:]

    bracketless_hierarchy_string = labelled_bracketless_hierarchy_string
    bracketless_hierarchy_labels = labelled_bracketless_hierarchy_labels

    max_label = max(bracketless_hierarchy_labels)

    colours = [
        "spring_green3",
        "cornflower_blue",
        "red3",
        "yellow2",
        "hot_pink",
        "pale_turquoise1",
    ]

    for counts, index in enumerate(string.ascii_lowercase):
        if counts % 2 == 0:
            myGlobals.console.print("")

        colour = colours[int(counts / 2.0)]

        # Create a new string of the same length as bracketless_hierarchy_string filled with spaces
        output_line = [" "] * len(bracketless_hierarchy_string)

        # Populate the output_line with characters from bracketless_hierarchy_string where the label matches the current index
        for i, (char, label, dim_string) in enumerate(
            zip(
                bracketless_hierarchy_string,
                bracketless_hierarchy_labels,
                labelled_dim_string,
            )
        ):
            if label == index:
                if dim_string == "d":
                    output_line[i] = f"[dim]{char}[/dim]"
                else:
                    if char == "[":
                        output_line[i] = r"\["
                    else:
                        output_line[i] = char

        line = "".join(output_line)
        myGlobals.console.print(f"[{colour}]{line}")
        myGlobals.my_decay_splash.append(f"[{colour}]{line}")

        if index == max_label:
            break

    myGlobals.console.print("")


def error_splash(stderr=None, pre_message="", post_message=""):
    print("\n\n")
    print(colored("++" * 50, "red"))
    if pre_message != "":
        print(f"ERROR {pre_message}\n")
    if stderr:
        if Path(stderr).is_file():
            f = open(stderr, "r")
            f_lines = f.readlines()
            for idx, line in enumerate(f_lines):
                line = line.replace("\n", "")
                print(colored(f"\t{line}", "red"))
    if post_message != "":
        print("\nERROR message:", colored(f"{post_message}", "red"))

    raise RuntimeError()


@contextmanager
def log_execution(message: str):
    myGlobals.console.print(
        f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] [dim] \t[deep_pink3]{message}...[deep_pink3]"
    )
    yield  # Allow the block of code to execute
    myGlobals.console.print(
        f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] [dim] \t[bold cyan][i]complete[/i][/bold cyan]"
    )


def info_print(message: str):
    myGlobals.console.print(
        f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] [dim] \t[grey39]{message}...[grey39]"
    )


def warning_print(message: str):
    myGlobals.console.print(
        f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] [dim] \t[red3]{message}...[red3]"
    )


@contextmanager
def status_execution(status_message: str, complete_message: str):
    myGlobals.console.print(
        f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] {status_message.replace('[bold green]', '[bold slate_blue3]')}"
    )
    with myGlobals.console.status(
        f"{status_message.replace('[bold green]', '[bold slate_blue3]')}",
        spinner="material",
    ):
        yield  # Allow the block of code to execute
    myGlobals.console.print(
        f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] \t{complete_message}"
    )


def events_table(events, tuple_loc):
    file = uproot.open(f"{tuple_loc}:DecayTree")

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Requested", justify="right", width=20)
    table.add_column("Returned", justify="right", width=20)
    table.add_column("Efficiency", justify="right", width=20)

    table.add_row(
        f"{events}",
        f"{file.num_entries}",
        f"{file.num_entries / events:.3f}",
    )

    myGlobals.console.print(table)


def print_file_info(file_path, time=None):
    # Get the file size in bytes
    size_bytes = os.path.getsize(file_path)

    file = uproot.open(f"{file_path}:DecayTree")

    # List of size units
    units = ["B", "KB", "MB", "GB", "TB", "PB"]
    unit_index = 0

    # Convert to the appropriate unit
    while size_bytes >= 1024 and unit_index < len(units) - 1:
        size_bytes /= 1024
        unit_index += 1

    # Print the file size with appropriate unit
    # print(f"File size: {size_bytes:.2f} {units[unit_index]}")
    myGlobals.console.print(
        f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] [sea_green1] Output file: [bold] {file_path} [bold] [sea_green1]"
    )
    myGlobals.console.print(
        f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] [sea_green1] [dim] \tsize: {size_bytes:.2f} {units[unit_index]} [sea_green1]"
    )
    myGlobals.console.print(
        f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] [sea_green1] [dim] \tentries: {file.num_entries} [sea_green1]"
    )
    if time:
        myGlobals.console.print(
            f"[grey39][{datetime.now().strftime('%H:%M:%S')}][grey39] [sea_green1] [dim] \tentries/min: {file.num_entries / (time / 60.0):.1f} [sea_green1]"
        )


def timings_table(only_rapidsim=False, only_vertexing=False):
    total_time = myGlobals.stopwatches.read_sum()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column(" ", style="bold", width=12)
    table.add_column("Configuration", justify="right", width=20)
    table.add_column("Generation", justify="right", width=20)
    table.add_column("Processing", justify="right", width=20)
    table.add_column("Total", justify="right", width=20)
    if not only_vertexing:
        table.add_row(
            "RapidSim",
            f"{myGlobals.stopwatches.read('RapidSim - config'):.2f}s ({myGlobals.stopwatches.read('RapidSim - config') / total_time * 100.0:.2f} %)",
            f"{myGlobals.stopwatches.read('RapidSim - generation'):.2f}s ({myGlobals.stopwatches.read('RapidSim - generation') / total_time * 100.0:.2f} %)",
            f"{myGlobals.stopwatches.read('RapidSim - processing'):.2f}s ({myGlobals.stopwatches.read('RapidSim - processing') / total_time * 100.0:.2f} %)",
            f"{myGlobals.stopwatches.read('RapidSim - config') + myGlobals.stopwatches.read('RapidSim - generation') + myGlobals.stopwatches.read('RapidSim - processing'):.2f}s ({(myGlobals.stopwatches.read('RapidSim - config') + myGlobals.stopwatches.read('RapidSim - generation') + myGlobals.stopwatches.read('RapidSim - processing')) / total_time * 100.0:.2f} %)",
        )
    if not only_rapidsim:
        table.add_row(
            "Vertexing",
            f"{myGlobals.stopwatches.read('Networks - config'):.2f}s ({myGlobals.stopwatches.read('Networks - config') / total_time * 100.0:.2f} %)",
            f"{myGlobals.stopwatches.read('Networks - generation'):.2f}s ({myGlobals.stopwatches.read('Networks - generation') / total_time * 100.0:.2f} %)",
            f"{myGlobals.stopwatches.read('Networks - processing'):.2f}s ({myGlobals.stopwatches.read('Networks - processing') / total_time * 100.0:.2f} %)",
            f"{myGlobals.stopwatches.read('Networks - config') + myGlobals.stopwatches.read('Networks - generation') + myGlobals.stopwatches.read('Networks - processing'):.2f}s ({(myGlobals.stopwatches.read('Networks - config') + myGlobals.stopwatches.read('Networks - generation') + myGlobals.stopwatches.read('Networks - processing')) / total_time * 100.0:.2f} %)",
        )
    if not only_vertexing and not only_rapidsim:
        table.add_row("", "", "")  # This will act as a horizontal line
        table.add_row(
            "Total",
            f"{myGlobals.stopwatches.read('RapidSim - config') + myGlobals.stopwatches.read('Networks - config'):.2f}s ({(myGlobals.stopwatches.read('RapidSim - config') + myGlobals.stopwatches.read('Networks - config')) / total_time * 100.0:.2f} %)",
            f"{myGlobals.stopwatches.read('RapidSim - generation') + myGlobals.stopwatches.read('Networks - generation'):.2f}s ({(myGlobals.stopwatches.read('RapidSim - generation') + myGlobals.stopwatches.read('Networks - generation')) / total_time * 100.0:.2f} %)",
            f"{myGlobals.stopwatches.read('RapidSim - processing') + myGlobals.stopwatches.read('Networks - processing'):.2f}s ({(myGlobals.stopwatches.read('RapidSim - processing') + myGlobals.stopwatches.read('Networks - processing')) / total_time * 100.0:.2f} %)",
            f"{total_time:.2f}s",
        )

    myGlobals.console.print(table)
    return total_time



import numpy as np
import re
import uproot
import os
import pandas as pd
import lhcb_rex.tools.display as display
import lhcb_rex.settings.globals as myGlobals


def find_closing_bracket_index(s, opening_index):
    # Check if the index is valid
    if s[opening_index] != "{":
        raise ValueError("The provided index does not point to an opening bracket.")

    stack = []

    for i in range(opening_index, len(s)):
        if s[i] == "{":
            stack.append(i)
        elif s[i] == "}":
            stack.pop()
            # If the stack is empty, we found the closing bracket
            if not stack:
                return i

    # If we finish the loop without finding a closing bracket
    return -1


def get_rapidsim_indicies(decay_string):
    decay_string = re.sub(r"\{(?!\s)", r"{ ", decay_string)
    decay_string = re.sub(r"(?<!\s)\}", r" }", decay_string)

    mother = decay_string.split("->")[0].replace(" ", "")

    daughter_string = decay_string[decay_string.find("->") + 2 :]
    if daughter_string[0] == " ":
        daughter_string = daughter_string[1:]

    labels = np.asarray([0 for i in daughter_string])

    for level in range(8):
        for idx, character in enumerate(daughter_string):
            if labels[idx] == level:
                if character == "{":
                    start_idx = idx
                    end_idx = find_closing_bracket_index(daughter_string, start_idx)
                    mother_in_level = daughter_string[start_idx + 1 : end_idx].split(
                        "->"
                    )[0]
                    labels[start_idx + 1 + len(mother_in_level) : end_idx] += (
                        1  # dont label the brackets # also bring mother down a level
                    )

    label_string = ""
    # convert labels into strings
    for idx, thing in enumerate(daughter_string):
        if thing == " ":
            label_string += " "
        else:
            label_string += str(labels[idx])

    daughter_split = daughter_string.split(" ")
    label_split = label_string.split(" ")

    ordering = np.empty((0, 2))
    for idx in range(len(daughter_split)):
        if daughter_split[idx] not in ["{", "}", "->", ""]:
            ordering = np.append(
                ordering, [[int(label_split[idx][0]), int(-1)]], axis=0
            )

    rapidsim_idx = 1  # mother is 0
    for level in range(8):
        for row in range(np.shape(ordering)[0]):
            if ordering[row][0] == level:
                ordering[row][1] = rapidsim_idx
                rapidsim_idx += 1

    # ordering[:,0] = order
    # ordering[:,1] = rapidsim index

    ordering_string = ""
    idx = 0
    for item in decay_string.split(" "):
        if item in ["{", "}", "->", ""]:
            ordering_string += item + " "
        elif item == mother:
            ordering_string += str(0) + " "
        else:
            ordering_string += str(int(ordering[idx][1])) + " "
            idx += 1
    ordering_string = ordering_string[:-1]  # remove last space

    stability_string = ""
    for idx in range(len(decay_string.split(" "))):
        if decay_string.split(" ")[idx] in ["{", "}", "->", ""]:
            stability_string += decay_string.split(" ")[idx] + " "
        else:
            if idx + 1 == len(decay_string.split(" ")):
                stability_string += "S" + " "  # stable, end of decay string
            elif decay_string.split(" ")[idx + 1] == "->":
                stability_string += "U" + " "
            else:
                stability_string += "S" + " "

    return (
        decay_string,
        ordering_string,
        stability_string,
        daughter_string,
        label_string,
    )


def compute_combined_param(daughter_info, param):
    daughters = list(daughter_info.keys())

    values = np.zeros(np.shape(daughter_info[daughters[0]]["M"]))
    if param in ["PX", "PY", "PZ"]:
        for daughter in daughters:
            values += daughter_info[daughter][param]
    elif param == "M":
        PE = values.copy()
        PX = values.copy()
        PY = values.copy()
        PZ = values.copy()

        for daughter in daughters:
            PE += np.sqrt(
                daughter_info[daughter]["M"] ** 2
                + daughter_info[daughter]["PX"] ** 2
                + daughter_info[daughter]["PY"] ** 2
                + daughter_info[daughter]["PZ"] ** 2
            )
            PX += daughter_info[daughter]["PX"]
            PY += daughter_info[daughter]["PY"]
            PZ += daughter_info[daughter]["PZ"]

        values = np.sqrt((PE**2 - PX**2 - PY**2 - PZ**2))

    return values


def organise_output_tuple(
    fileName,
    dropMissing,
    naming_scheme,
    mass_hypotheses,
    combined_particles,
    true_PID_scheme,
):
    # Open the input file and access the TTree
    with uproot.open(fileName) as f:
        # Get the tree
        tree = f["DecayTree"]

        # Load tree into numpy
        if dropMissing:
            branches_to_keep = []
            for particle in naming_scheme.values():
                branches_to_keep.extend(
                    [
                        branch
                        for branch in tree.keys()
                        if branch[: len(particle)] == particle
                    ]
                )
        else:
            branches_to_keep = tree.keys()

        map_NA_codes = {}
        idx = 0
        for branch in branches_to_keep:
            if "origX_TRUE" in branch:
                if (
                    re.match(r"^NA_\d{8}$", naming_scheme[idx])
                    or naming_scheme[idx] == "NA"
                ):
                    map_NA_codes[naming_scheme[idx]] = branch.replace("_origX_TRUE", "")
                idx += 1

        branches = {
            branch: tree[branch].array(library="np") for branch in branches_to_keep
        }

        # Reset mass hypotheses
        if mass_hypotheses:
            for mass_hypothesis in mass_hypotheses:
                particle = mass_hypotheses[mass_hypothesis]
                filtered_df = myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                    (
                        myGlobals.rapidsim_settings["RAPIDSIM_particles_df"]["part"]
                        == particle
                    )
                    | (
                        myGlobals.rapidsim_settings["RAPIDSIM_particles_df"]["anti"]
                        == particle
                    )
                ]
                new_mass = float(filtered_df.mass)
                branches[f"{mass_hypothesis}_M"] = (
                    np.ones(np.shape(branches[f"{mass_hypothesis}_M"])) * new_mass
                )

        # Recompute variables of reconstructed particles
        if combined_particles:
            for combined_particle in combined_particles:
                particle_daughters = combined_particles[combined_particle]
                daughter_info = {}
                for daughter in particle_daughters:
                    daughter_info[daughter] = {}
                    for param in ["M", "PX", "PY", "PZ"]:
                        daughter_info[daughter][param] = branches[f"{daughter}_{param}"]
                for param in ["M", "PX", "PY", "PZ"]:
                    branches[f"{combined_particle}_{param}"] = compute_combined_param(
                        daughter_info, param
                    )
                branches[f"{combined_particle}_P"] = np.sqrt(
                    branches[f"{combined_particle}_PX"] ** 2
                    + branches[f"{combined_particle}_PY"] ** 2
                    + branches[f"{combined_particle}_PZ"] ** 2
                )
                branches[f"{combined_particle}_PT"] = np.sqrt(
                    branches[f"{combined_particle}_PX"] ** 2
                    + branches[f"{combined_particle}_PY"] ** 2
                )

        # True PID
        for particle in true_PID_scheme:
            if f"{particle}_PX" in list(branches.keys()):
                branches[f"{particle}_ID_TRUE"] = (
                    np.ones(np.shape(branches[f"{particle}_PX"]))
                    * true_PID_scheme[particle]
                )
                # maybe only want to write this out for final state particles?
                branches[f"{particle}_ID"] = (
                    np.ones(np.shape(branches[f"{particle}_PX"]))
                    * true_PID_scheme[particle]
                )
            if particle in map_NA_codes:
                if f"{map_NA_codes[particle]}_PX" in list(branches.keys()):
                    branches[f"{map_NA_codes[particle]}_ID_TRUE"] = (
                        np.ones(np.shape(branches[f"{map_NA_codes[particle]}_PX"]))
                        * true_PID_scheme[particle]
                    )

        # PID from mass hypotheses
        if mass_hypotheses:
            for particle in mass_hypotheses:
                if (
                    myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                        (
                            myGlobals.rapidsim_settings["RAPIDSIM_particles_df"]["part"]
                            == mass_hypotheses[particle]
                        )
                    ].shape[0]
                    == 1
                ):
                    pdg = int(
                        myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                            (
                                myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                                    "part"
                                ]
                                == mass_hypotheses[particle]
                            )
                        ].ID
                    )
                elif (
                    myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                        (
                            myGlobals.rapidsim_settings["RAPIDSIM_particles_df"]["anti"]
                            == mass_hypotheses[particle]
                        )
                    ].shape[0]
                    == 1
                ):
                    pdg = int(
                        myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                            (
                                myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                                    "anti"
                                ]
                                == mass_hypotheses[particle]
                            )
                        ].ID
                        * -1
                    )
                else:
                    display.error_splash(
                        post_message=f"{particle} not found in {myGlobals.rapidsim_settings['RAPIDSIM_ROOT']}/config/particles.dat, or listed multiple times."
                    )
                branches[f"{particle}_ID"] = (
                    np.ones(np.shape(branches[f"{particle}_PX"])) * pdg
                )

        # Reorder the columns alphabetically
        branches = pd.DataFrame(branches)
        branches = branches[sorted(branches.columns)]
        branches = {col: branches[col].to_numpy() for col in branches.columns}

        # Open the output file and write the new tree with the selected branches
        with uproot.recreate(fileName.replace(".root", "_temp.root")) as new_file:
            new_file["DecayTree"] = branches

    os.system(f"mv {fileName.replace('.root', '_temp.root')} {fileName}")

    return map_NA_codes


def check_params(line, checked_lines, key, params_list):
    var_list = line[line.find(":") + 1 :].replace(",", "").replace("\n", "").split(" ")
    try:
        var_list.remove("")
    except Exception:
        pass
    var_list.extend(params_list)
    var_list = np.unique(var_list)

    line = line[: line.find(":") + 1]
    for var in var_list:
        line += f" {var},"
    line = line[:-1] + "\n"  # remove last comma
    checked_lines[key] = True
    return line, checked_lines


def check_line(checked_lines, option, choice):
    line = f"{option} : {choice}\n"
    checked_lines[option] = True
    return line, checked_lines


def write_line(checked_lines, option, choice, f):
    line = f"{option} : {choice}\n"
    checked_lines[option] = True
    f.write(line)
    return checked_lines


def check_config(
    config,
    USE_EVTGEN,
    naming_scheme,
    model_scheme=None,
    geometry="LHCb",
    acceptance="AllIn",
    useEvtGen="TRUE",
    evtGenUsePHOTOS="TRUE",
    dropMissing=True,
):
    blocked_settings = [
        "paramsTwoBody",
        "paramsThreeBody",
        "paramsFourBody",
        "paramsFiveBody",
        "paramsSixBody",
    ]  # temporarily blocking this function in rapidsim

    settings = {}
    settings[-1] = {}
    for key in naming_scheme:
        settings[int(key)] = {}

    particle_index = 0
    current_particle = -1

    f = open(config, "r")
    f_lines = f.readlines()
    for idx, line in enumerate(f_lines):
        if ":" in line:
            option = line.split(":")[0].replace(" ", "").replace("\t", "")
            setting = (
                line.split(":")[-1]
                .replace(" ", "")
                .replace("\n", "")
                .replace(",", ", ")
            )
            settings[current_particle][option] = setting
        elif f"@{particle_index}" in line:
            current_particle = particle_index
            particle_index += 1

    # particle names
    for key in naming_scheme:
        # if naming_scheme[key] != 'NA':
        if (
            not re.match(r"^NA_\d{8}$", naming_scheme[key])
            and naming_scheme[key] != "NA"
        ):
            settings[int(key)]["name"] = naming_scheme[key]
        else:
            settings[int(key)]["invisible"] = (
                "true"  # dont include these in mother mass variables
            )

    # EVTGEN models
    for key in model_scheme:
        if not USE_EVTGEN and model_scheme[key] != "PHSP":
            print(
                f"Cannot use {model_scheme[key]}, $EVTGEN_ROOT not set. Using PHSP instead."
            )
            settings[int(key)]["evtGenModel"] = "PHSP"
        else:
            settings[int(key)]["evtGenModel"] = model_scheme[key]

    # geometry
    settings[-1]["geometry"] = geometry

    # acceptance
    settings[-1]["acceptance"] = acceptance

    # useEvtGen
    if not USE_EVTGEN and useEvtGen == "TRUE":
        print("Cannot set useEvtGen : TRUE, $EVTGEN_ROOT not set.")
    else:
        settings[-1]["useEvtGen"] = useEvtGen

    # evtGenUsePHOTOS
    if evtGenUsePHOTOS != "TRUE":
        pass  # dont set as setting option to anything sets photos.
    elif not USE_EVTGEN and evtGenUsePHOTOS == "TRUE":
        print("Cannot set evtGenUsePHOTOS : TRUE, $EVTGEN_ROOT not set.")
    else:
        settings[-1]["evtGenUsePHOTOS"] = evtGenUsePHOTOS

    paramsDecaying = [
        "M",
        "P",
        "PT",
        "PX",
        "PY",
        "PZ",
        "vtxX",
        "vtxY",
        "vtxZ",
        "origX",
        "origY",
        "origZ",
    ]
    paramsStable = ["M", "P", "PT", "PX", "PY", "PZ", "origX", "origY", "origZ"]

    # paramsStable
    if "paramsStable" in settings[-1].keys():
        settings[-1]["paramsStable"] = (
            settings[-1]["paramsStable"].replace(" ", "").split(",")
        )
        settings[-1]["paramsStable"].extend(paramsStable)
        settings[-1]["paramsStable"] = np.unique(settings[-1]["paramsStable"])
        line = ""
        for var in settings[-1]["paramsStable"]:
            line += f" {var},"
        line = line[1:-1] + "\n"  # remove last comma and initial space
        settings[-1]["paramsStable"] = line
    else:
        settings[-1]["paramsStable"] = paramsStable

    # paramsDecaying
    if "paramsDecaying" in settings[-1].keys():
        settings[-1]["paramsDecaying"] = (
            settings[-1]["paramsDecaying"].replace(" ", "").split(",")
        )
        settings[-1]["paramsDecaying"].extend(paramsDecaying)
        settings[-1]["paramsDecaying"] = np.unique(settings[-1]["paramsDecaying"])
        line = ""
        for var in settings[-1]["paramsDecaying"]:
            line += f" {var},"
        line = line[1:-1] + "\n"  # remove last comma and initial space
        settings[-1]["paramsDecaying"] = line
    else:
        settings[-1]["paramsDecaying"] = paramsDecaying

    with open(config.replace(".config", "_temp.config"), "w") as f_out:
        for key in [-1] + list(range(np.amax(list(naming_scheme.keys())) + 1)):
            if key != -1:
                line = f"@{int(key)}\n"
                f_out.write(line)

            for option in list(settings[int(key)].keys()):
                setting = settings[int(key)][option]
                if key != -1:
                    if option == "evtGenModel":
                        line = f"\t{option} : {setting.replace('_', ' ')}\n"  # replace _ in evtgenmodel name
                    else:
                        line = f"\t{option} : {setting}\n"
                else:
                    if option not in blocked_settings:
                        line = f"{option} : {setting}\n"

                f_out.write(line)

    os.system(f"mv {config.replace('.config', '_temp.config')} {config}")

    return settings



import os
from datetime import datetime
import os.path
from os import path

import importlib.resources


def mkdir(save_dir, allow_exists=False):
    if allow_exists:
        os.makedirs(save_dir, exist_ok=True)
    else:
        os.mkdir(save_dir)


def copy(file, location):
    location = f"{location}/."
    try:
        os.system(f"cp {file} {location}")
    except Exception:
        pass


def submit(
    job_name,
    numJobs,
    save_directory_base,
    options,
    clean_log=True,
    overwrite=False,
    memory=12,
    hours=5,
    DRY=True,
    log_dir=None,
    CODEPATH="/users/am13743/Rex/",
    EXE="src/lhcb_rex/data/batch/make_graphs_batch.py",
):
    if log_dir is not None:
        if clean_log:
            os.system(f"rm -r {log_dir}/*{job_name}*")
        mkdir(log_dir, allow_exists=True)

    save_directory_base = save_directory_base.replace("//", "/")
    save_directory = save_directory_base + job_name

    if path.exists(save_directory):
        if overwrite:
            print(f"{save_directory} exists, deleting content...")
            os.system(f"rm -r {save_directory}")
        else:
            print(save_directory, "exists... quitting....")
            quit()

    mkdir(save_directory)

    print("\nOutput to be saved to", save_directory)

    # Setting up submission directory
    now = datetime.now()
    dt_string = now.strftime("%Hhr_%Mmin_%Ssec_%d_%m_%Y")
    tag_directory = f"{save_directory}/{job_name}_{dt_string}"
    mkdir(tag_directory)

    # job_options = [
    #     "--processID $1",
    # ]
    # job_options = " ".join(job_options)
    # # options_string = ""
    # for key in options:
    #     job_options += f" --{key} {options[key]}"

    job_options = ["--processID $1"]
    for key, value in options.items():
        if isinstance(value, str):
            job_options.append(f'--{key} "{value}"')
        else:
            job_options.append(f"--{key} {value}")

    job_options_str = " ".join(job_options)
    job_options = job_options_str
    # print(job_options)
    # # os.system(f"python ../make_graphs_batch.py{options_string}")
    # quit()

    os.mkdir(f"temp{job_name}")

    # ENVPATH = "/users/am13743/fast_vertexing_variables/training/condor/batch/"

    # Creating run file from template
    f = importlib.resources.files("lhcb_rex").joinpath("data/batch/batch.sh").open("r")
    f_lines = f.readlines()
    with open(f"temp{job_name}/batch.sh", "a") as f_out:
        for idx, line in enumerate(f_lines):
            # if "ENVPATH" in line:
            #     line = line.replace("ENVPATH", ENVPATH)

            if "CODEPATH" in line:
                line = line.replace("CODEPATH", CODEPATH)

            if "PYTHONOPTIONS" in line:
                line = line.replace("PYTHONOPTIONS", job_options)

            if "EXE" in line:
                line = line.replace("EXE", EXE)

            f_out.write(line)

    os.system(f"chmod +x temp{job_name}/batch.sh")

    # Creating submit file from template
    f = (
        importlib.resources.files("lhcb_rex")
        .joinpath("data/batch/submit.job")
        .open("r")
    )
    f_lines = f.readlines()
    with open(f"temp{job_name}/submit.job", "a") as f_out:
        for idx, line in enumerate(f_lines):
            if "JOBNAME" in line:
                line = line.replace("JOBNAME", job_name)

            if "NUMJOBS" in line:
                line = line.replace("NUMJOBS", str(numJobs))

            if "LOGDIR" in line:
                if log_dir is not None:
                    line = line.replace("LOGDIR", log_dir)

            if "TIME" in line:
                line = line.replace("TIME", str(60 * 60 * hours))

            if "EXE" in line:
                line = line.replace("EXE", "batch.sh")

            if "ARGS" in line:
                line = line.replace("ARGS", "$(PROCESS)")

            if "MEMORY" in line:
                line = line.replace("MEMORY", str(memory))

            if "log = " in line or "output = " in line or "error = " in line:
                if log_dir is not None:
                    f_out.write(line)
            else:
                f_out.write(line)

    copy(f"temp{job_name}/submit.job", save_directory)
    copy(f"temp{job_name}/batch.sh", save_directory)

    # Submitting jobs
    print(f"\n\nSubmitting {numJobs} jobs, job_name: {job_name} ...\n")
    pwd = os.getcwd()
    os.chdir(f"{save_directory}/")
    if not DRY:
        os.system("condor_submit submit.job")
    else:
        print("not executing...", "condor_submit submit.job")
        print("\n\n")
        os.system("cat submit.job")
        print("\n\n")
        os.system("cat batch.sh")
        print("\n\n")

    os.chdir(pwd)
    os.system(f"rm -r temp{job_name}/")



import lhcb_rex.settings.globals as myGlobals
from lhcb_rex.data.data_manager import tuple_manager
from lhcb_rex.inference.inference_network_manager import Network
from lhcb_rex.inference.inference_graph_constructor import GraphConstructor
import lhcb_rex.tools.display as display
import lhcb_rex.tools.variables_tools as pts
import numpy as np
import pandas as pd
import warnings
import lhcb_rex.tools.get_weights as get_weights


warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
warnings.simplefilter(action="ignore", category=RuntimeWarning)
warnings.simplefilter(action="ignore", category=UserWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)


def run_inference_from_tuple(
    tuple_location,
    mother_particle,
    daughter_particles,
    intermediate_particles=None,
    branch_naming_structure=None,
    reconstruction_topology=None,
    physical_units="GeV",
    mass_hypotheses=None,
    stages=["smearPV", "smearMomenta", "genPID", "runVertexing"],
):
    print(reconstruction_topology)

    hf_model_paths = get_weights.get_model_paths()

    if reconstruction_topology is None:
        daughters_str = " ".join(daughter_particles)
        reconstruction_topology = f"{mother_particle} -> {daughters_str}"
    graph_structure = GraphConstructor(reconstruction_topology)
    # print('plotting', reconstruction_topology)
    # graph_structure.plot()
    intermediate_particles = graph_structure.intermediate_particles

    new_branches_to_keep = []

    myGlobals.particle_map = {}  # 2 two way map to translate between particle names
    myGlobals.particle_map["toDAUGHTERS"] = {}
    myGlobals.particle_map["fromDAUGHTERS"] = {}
    for idx, daughter_particle in enumerate(daughter_particles):
        myGlobals.particle_map["fromDAUGHTERS"][f"DAUGHTER{idx + 1}"] = (
            daughter_particle
        )
        myGlobals.particle_map["toDAUGHTERS"][daughter_particle] = f"DAUGHTER{idx + 1}"
    myGlobals.particle_map["toINTERMEDIATES"] = {}
    myGlobals.particle_map["fromINTERMEDIATES"] = {}
    if intermediate_particles is not None:
        for idx, intermediate_particle in enumerate(intermediate_particles):
            # ridx = random.randint(1000000, 9999999)
            myGlobals.particle_map["fromINTERMEDIATES"][f"INTERMEDIATE{idx}"] = (
                intermediate_particle
            )
            myGlobals.particle_map["toINTERMEDIATES"][intermediate_particle] = (
                f"INTERMEDIATE{idx}"
            )

    Nparticles = len(list(myGlobals.particle_map["fromDAUGHTERS"].keys()))

    if physical_units == "GeV":
        myGlobals.masses = myGlobals.masses_GeV
        myGlobals.particle_masses_dict_mothers = (
            myGlobals.particle_masses_dict_mothers_GeV
        )

    combined_particles = {}
    combined_particles["MOTHER"] = list(myGlobals.particle_map["fromDAUGHTERS"].keys())
    if intermediate_particles is not None:
        for intermediate_particle, daughters in intermediate_particles.items():
            combined_particles[
                myGlobals.particle_map["toINTERMEDIATES"][intermediate_particle]
            ] = [
                myGlobals.particle_map["toDAUGHTERS"][daughter]
                for daughter in daughters
            ]

    with (
        myGlobals.stopwatches.timer("Networks - processing"),
        display.status_execution(
            status_message="[bold green]Staging tuple...",
            complete_message="[bold green]Tuple staged :white_check_mark:",
        ),
        display.log_execution("Reading tuple"),
    ):
        data_tuple = tuple_manager(
            tuple_location=tuple_location,
            mother_particle_name=mother_particle,
            daughter_particle_names=daughter_particles,
            combined_particles=combined_particles,
            branch_naming_structure=branch_naming_structure,
            mass_hypotheses=mass_hypotheses,
        )
    # run some pytests for this function

    with (
        myGlobals.stopwatches.timer("Networks - config"),
        display.status_execution(
            status_message="[bold green]Initialising networks...",
            complete_message="[bold green]Networks initialised :white_check_mark:",
        ),
    ):
        if "smearPV" in stages:
            with display.log_execution("Initialising PV smearing network"):
                PVModel = Network(
                    network="PV_smear_diffusion",
                    model_parameters="PVsmear_model_parameters.json",
                    # pkl="/users/am13743/Rex/logs/July_pv_smear/version_23/models.pkl",  # normal
                    pkl=hf_model_paths["smearPV"],
                    Nparticles=Nparticles,
                    physical_units=physical_units,
                    daughter_particle_names=list(daughter_particles),
                )

        if "smearMomenta" in stages:
            with display.log_execution("Initialising momentum smearing network"):
                smearModel = Network(
                    network="mom_smear_diffusion",
                    model_parameters="diffusion_model_parameters.json",
                    # pkl="/users/am13743/Rex/logs/July_smear_diffusion2/version_0/models.pkl",
                    pkl=hf_model_paths["smearMomenta"],
                    N_diffusion_steps=50,
                    # N_diffusion_steps=1,
                    Nparticles=Nparticles,
                    physical_units=physical_units,
                    daughter_particle_names=list(daughter_particles),
                    EMA=True,
                )

        if "genPID" in stages:
            with display.log_execution("Initialising PID network"):
                PIDModel = Network(
                    network="PID_trig_diffusion",
                    model_parameters="diffusion_model_parameters.json",
                    # pkl="/users/am13743/Rex/logs/PID_diffusion/version_0/models.pkl",
                    pkl=hf_model_paths["genPID"],
                    N_diffusion_steps=50,
                    # N_diffusion_steps=1,
                    Nparticles=Nparticles,
                    physical_units=physical_units,
                    daughter_particle_names=list(daughter_particles),
                    EMA=False,
                )

        if "runVertexing" in stages:
            with display.log_execution("Initialising vertexing network"):
                vtxModel = Network(
                    network="reco_vertex_diffusion",
                    model_parameters="diffusion_model_parameters.json",
                    # pkl="/users/am13743/Rex/logs/July_vertex_diffusion_w_encoder_IPDIRA/version_25/demo.pkl", # A
                    # pkl="/users/am13743/Rex/logs/July_vertex_diffusion_w_encoder_IPDIRA/version_25/models.pkl",  # A
                    # pkl="/users/am13743/Rex/logs/July_vertex_diffusion_w_encoder_IPDIRA_adddeltas/version_0/demo.pkl", # A


                    # pkl="/users/am13743/Rex/logs/July_vertex_diffusion_w_encoder_IPDIRA_adddeltas/version_2/models_20_2.pkl", # A
                    pkl="/users/am13743/Rex/logs/July_vertex_diffusion_w_encoder_IPDIRA_adddeltas/version_2/models_22_1.pkl", # A
                    # pkl="/users/am13743/Rex/logs/July_vertex_diffusion_w_encoder_IPDIRA_adddeltas/version_0/demo.pkl", # A
                    # pkl="/users/am13743/Rex/logs/July_vertex_diffusion_w_encoder_IPDIRA_adddeltas/version_1/models copy 2.pkl", # A

                    # pkl=hf_model_paths["runVertexing"],
                    N_diffusion_steps=25,
                    # N_diffusion_steps=1,
                    Nparticles=Nparticles,
                    physical_units=physical_units,
                    daughter_particle_names=list(daughter_particles),
                    EMA=False,
                )

    with myGlobals.stopwatches.timer("Networks - processing"):
        data_tuple.append_initial_conditional_information(data_tuple.true_PID_scheme)
        data_tuple.append_mother_flight()

    if "smearPV" in stages:
        with (
            myGlobals.stopwatches.timer("Networks - generation"),
            display.status_execution(
                status_message="[bold green]Smearing primary vertex...",
                complete_message="[bold green]Primary vertex smeared :white_check_mark:",
            ),
        ):
            smearingPV_conditions = data_tuple.get_branches(
                PVModel.conditions,
                PVModel.transformers,
                numpy=True,
            )

            smeared_PV_output = PVModel.query_network(
                smearingPV_conditions,
            )

            data_tuple.smearPV(smeared_PV_output)

    if "smearMomenta" in stages:
        with (
            myGlobals.stopwatches.timer("Networks - generation"),
            display.status_execution(
                status_message="[bold green]Smearing momenta...",
                complete_message="[bold green]Momenta smeared :white_check_mark:",
            ),
        ):
            smearing_node_conditions = data_tuple.get_branches(
                smearModel.conditions,
                smearModel.transformers,
                numpy=True,
            )

            smearing_edge_conditions = data_tuple.get_branches(
                smearModel.edge_conditions,
                smearModel.edge_transformers,
                numpy=False,
            )

            myGlobals.personalised_track_node_types = True
            mom_smearing_output = smearModel.query_network(
                data_tuple.true_PID_scheme,
                smearing_node_conditions,
                smearing_edge_conditions,
            )
            myGlobals.personalised_track_node_types = False

            data_tuple.propagate_smearing(mom_smearing_output, Nparticles=Nparticles)
            data_tuple.recompute_combined_particles(
                combined_particles, list(myGlobals.particle_map["fromDAUGHTERS"].keys())
            )
            new_branches_to_keep.extend([i for i in mom_smearing_output.keys() if "delta" not in i])

    with myGlobals.stopwatches.timer("Networks - processing"):
        data_tuple.append_secondary_conditional_information()

    if "genPID" in stages:
        with (
            myGlobals.stopwatches.timer("Networks - generation"),
            display.status_execution(
                status_message="[bold green]Applying PID and triger...",
                complete_message="[bold green]PID and triger applied :white_check_mark:",
            ),
        ):
            pid_node_conditions = data_tuple.get_branches(
                PIDModel.conditions,
                PIDModel.transformers,
                numpy=True,
            )

            pid_edge_conditions = data_tuple.get_branches(
                PIDModel.edge_conditions,
                PIDModel.edge_transformers,
                numpy=False,
            )

            myGlobals.personalised_track_node_types = True
            pid_trigger_output = PIDModel.query_network(
                data_tuple.true_PID_scheme, pid_node_conditions, pid_edge_conditions
            )
            myGlobals.personalised_track_node_types = False

            data_tuple.add_branches(pid_trigger_output)
            new_branches_to_keep.extend(list(pid_trigger_output.keys()))

    with myGlobals.stopwatches.timer("Networks - processing"):
        data_tuple.append_tertiary_conditional_information(
            combined_particles,
            data_tuple.true_PID_scheme,
            list(myGlobals.particle_map["fromDAUGHTERS"].keys()),
        )

    if "runVertexing" in stages:
        with (
            myGlobals.stopwatches.timer("Networks - generation"),
            display.status_execution(
                status_message="[bold green]Running vertexing...",
                complete_message="[bold green]Vertexing complete :white_check_mark:",
            ),
        ):
            vtx_mother_conditions = data_tuple.get_branches(
                vtxModel.mother_conditions,
                vtxModel.transformers,
                numpy=False,
            )
            vtx_intermediate_conditions = data_tuple.get_branches(
                vtxModel.intermediate_conditions,
                vtxModel.transformers,
                numpy=False,
            )
            vtx_track_conditions = data_tuple.get_branches(
                vtxModel.track_conditions,
                vtxModel.transformers,
                numpy=False,
            )

            vtx_mother_conditions = np.asarray(
                vtx_mother_conditions[vtxModel.mother_conditions]
            )
            vtx_intermediate_conditions = np.asarray(
                vtx_intermediate_conditions[vtxModel.intermediate_conditions]
            )
            vtx_track_conditions = np.asarray(
                vtx_track_conditions[vtxModel.track_conditions]
            )

            vtx_track_conditions = np.reshape(
                vtx_track_conditions, (-1, Nparticles, len(myGlobals.track_conditions))
            )

            out_df = vtxModel.query_network_vtx(
                graph_structure,
                vtx_mother_conditions,
                vtx_intermediate_conditions,
                vtx_track_conditions,
                Nparticles=Nparticles,
            )

            data_tuple.add_branches(out_df)
            new_branches_to_keep.extend(list(out_df.keys()))

    with myGlobals.stopwatches.timer("Networks - processing"):
        # data_tuple.write(
        # new_branches_to_keep,
        # output_location=f"{data_tuple.tuple_location[:-5]}_reco.root",
        # keep_vertex_info=True,
        # keep_tuple_structure=False,
        # extra_branches=[],)

        # quit()

        new_branches_to_keep_user_naming = []

        new_branches = []
        temp_particles = {}
        for idx, particle in enumerate(
            list(myGlobals.particle_map["toDAUGHTERS"].keys())
        ):
            temp_particles[particle] = str(np.random.randint(0, 99999999))
        temp_combined_particles = {}
        for idx, combined_particle in enumerate(combined_particles):
            if combined_particle != "MOTHER":
                temp_combined_particles[
                    myGlobals.particle_map["fromINTERMEDIATES"][combined_particle]
                ] = str(np.random.randint(0, 99999999))

        for branch in data_tuple.tuple:
            branch_in = branch
            for combined_particle in combined_particles:
                if combined_particle == "MOTHER":
                    branch = branch.replace(combined_particle, mother_particle)
                else:
                    branch = branch.replace(
                        combined_particle,
                        temp_combined_particles[
                            myGlobals.particle_map["fromINTERMEDIATES"][
                                combined_particle
                            ]
                        ],
                    )

            for combined_particle in combined_particles:
                if combined_particle != "MOTHER":
                    branch = branch.replace(
                        temp_combined_particles[
                            myGlobals.particle_map["fromINTERMEDIATES"][
                                combined_particle
                            ]
                        ],
                        myGlobals.particle_map["fromINTERMEDIATES"][combined_particle],
                    )

            for particle in list(myGlobals.particle_map["toDAUGHTERS"].keys()):
                branch = branch.replace(
                    myGlobals.particle_map["toDAUGHTERS"][particle],
                    temp_particles[particle],
                )
            for particle in list(myGlobals.particle_map["toDAUGHTERS"].keys()):
                branch = branch.replace(temp_particles[particle], particle)
            new_branches.append(branch)
            if branch_in in new_branches_to_keep:
                new_branches_to_keep_user_naming.append(branch)

        data_tuple.tuple.columns = new_branches

        data_tuple.tuple = data_tuple.tuple[
            data_tuple.original_branches + new_branches_to_keep_user_naming
        ]

        # re-order columns
        columns = data_tuple.tuple.columns
        prefix_order = [mother_particle]
        for intermediate in combined_particles:
            if intermediate != mother_particle:
                prefix_order.append(
                    myGlobals.particle_map["fromINTERMEDIATES"][intermediate]
                )
        for daughter_particle in daughter_particles:
            prefix_order.append(daughter_particle)

        ordered_columns = []
        for prefix in prefix_order:
            cols_with_prefix = [col for col in columns if col.startswith(prefix)]
            ordered_columns.extend(cols_with_prefix)
        data_tuple.tuple = data_tuple.tuple[ordered_columns]

        output_location = f"{data_tuple.tuple_location[:-5]}_reco.root"

        pts.write_df_to_root(data_tuple.tuple, output_location, "DecayTree")

    print(f"Saving every branch to {output_location}")

    return output_location


import lhcb_rex.data.graph_maker as graph_maker
from torch_geometric.loader import DataLoader
from torch.utils.data import Subset
from pydantic import BaseModel, Field
from typing import Dict
from rich import print
from torch.utils.data import WeightedRandomSampler
import os
import torch


class SettingConfig(BaseModel):
    file_name: str
    splits: int = Field(gt=0, description="Number of splits, must be > 0")
    batch_size: int = None


def create_settings(file: str, splits: int, batch_size: int = None) -> SettingConfig:
    """Helper function to create a settings configuration."""
    return SettingConfig(
        file_name=file,
        splits=splits,
        batch_size=batch_size,
    )


class DataLoaderSettings(BaseModel):
    settings: Dict[str, SettingConfig]


def get_loaders(
    settings: DataLoaderSettings,
    batch_size: int = 256,
    batch_size_val: int = -1,
    verbose: bool = False,
    use_weights: bool = False,
    load_physics_validation: bool = False,
    enhance_fully_reco=1.0,
    extra_val_data_locs=None,
):
    if batch_size_val == -1:
        batch_size_val = batch_size

    dataloaders = {}
    val_dataloaders = {}

    training_info = {}
    training_info["path"] = []
    training_info["particles_involved"] = []

    for setting in settings:
        root = settings[setting].file_name

        if settings[setting].batch_size:
            batch_size_to_use = settings[setting].batch_size
        else:
            batch_size_to_use = batch_size

        dataset = graph_maker.GENERALISED_HeteroGraphDataset(
            # config='/users/am13743/Rex/davinci/configs/configs_2_body/config_dict_B0_AAA+BBB-.pickle',
            root=root,
            mode="train",
            # max_N_samples_per=-1,
            splits=settings[setting].splits,
            use_weights=use_weights,
        )

        fully_reco_enhancement = [
            enhance_fully_reco if x == 1.0 else 1.0
            for x in dataset.options["fully_reco_bool"]
        ]
        dataloader = DataLoader(dataset, batch_size=batch_size_to_use, shuffle=True)
        if use_weights:
            weights = dataset.options["training_weights"]
            if enhance_fully_reco > 1.0:
                weights = torch.multiply(weights, fully_reco_enhancement)
            sampler = WeightedRandomSampler(
                weights, num_samples=len(weights), replacement=True
            )
            dataloader = DataLoader(
                dataset, batch_size=batch_size_to_use, sampler=sampler
            )
        else:
            if enhance_fully_reco > 1.0:
                sampler = WeightedRandomSampler(
                    fully_reco_enhancement,
                    num_samples=len(fully_reco_enhancement),
                    replacement=True,
                )
                dataloader = DataLoader(
                    dataset, batch_size=batch_size_to_use, sampler=sampler
                )
            else:
                dataloader = DataLoader(
                    dataset, batch_size=batch_size_to_use, shuffle=True
                )

        dataloaders[setting] = dataloader

        branch_options = dataset.options

        if branch_options != dataset.options:
            raise ValueError(
                "Attempt was made to load datasets with differing options/branches."
            )

        # print(branch_options.keys())
        # print(branch_options['settings'])
        # quit()

        training_info["path"].append(branch_options["path"])
        # training_info["particles_involved"].append(branch_options["particles_involved"])
        training_info["particles_involved"] = [
            i for i in branch_options["settings"]["branches"].keys() if "DAUGHTER" in i
        ]

        print(root)
        dataset = graph_maker.GENERALISED_HeteroGraphDataset(
            root=root,
            mode="test",
            # max_N_samples_per=-1,
            splits=settings[setting].splits,
            use_weights=False,
        )
        dataloader = DataLoader(dataset, batch_size=batch_size_val, shuffle=False)
        val_dataloaders[setting] = dataloader

    if extra_val_data_locs:
        for setting in extra_val_data_locs:
            dataset = graph_maker.GENERALISED_HeteroGraphDataset(
                root=extra_val_data_locs[setting]["loc"],
                mode="test",
                splits=1,
                use_weights=False,
            )
            dataloader = DataLoader(dataset, batch_size=256, shuffle=False)
            val_dataloaders[setting] = dataloader

    if load_physics_validation:
        root_dir = os.path.dirname(root)
        dataset = graph_maker.HeteroGraphDataset(
            root=f"{root_dir}/graphs_3_int_fullrecosequentialValidationKee_PID",
            mode="train",
            max_N_samples_per=-1,
            splits=1,
            use_weights=False,
        )
        subset_dataset = Subset(dataset, range(min(25000, len(dataset))))
        dataloader = DataLoader(subset_dataset, batch_size=256, shuffle=False)
        val_dataloaders["physics_Kee_3"] = dataloader

        root_dir = os.path.dirname(root)
        dataset = graph_maker.HeteroGraphDataset(
            root=f"{root_dir}/graphs_3_int_fullrecosequentialValidationKstee_PID",
            mode="train",
            max_N_samples_per=-1,
            splits=1,
            use_weights=False,
        )
        subset_dataset = Subset(dataset, range(min(25000, len(dataset))))
        dataloader = DataLoader(subset_dataset, batch_size=256, shuffle=False)
        val_dataloaders["physics_Kstee_3"] = dataloader

        root_dir = os.path.dirname(root)
        dataset = graph_maker.HeteroGraphDataset(
            root=f"{root_dir}/graphs_3_int_fullrecosequentialValidationBuD0piKenu_PID",
            mode="train",
            max_N_samples_per=-1,
            splits=1,
            use_weights=False,
        )
        subset_dataset = Subset(dataset, range(min(25000, len(dataset))))
        dataloader = DataLoader(subset_dataset, batch_size=256, shuffle=False)
        val_dataloaders["physics_BuD0piKenu_3"] = dataloader

        root_dir = os.path.dirname(root)
        dataset = graph_maker.HeteroGraphDataset(
            root=f"{root_dir}/graphs_3_int_fullrecosequentialValidationKmumu_PID",
            mode="train",
            max_N_samples_per=-1,
            splits=1,
            use_weights=False,
        )
        subset_dataset = Subset(dataset, range(min(25000, len(dataset))))
        dataloader = DataLoader(subset_dataset, batch_size=256, shuffle=False)
        val_dataloaders["physics_Kmumu_3"] = dataloader

        # root_dir = os.path.dirname(root)
        # dataset = graph_maker.HeteroGraphDataset(
        #     root=f'{root_dir}/graphs_3_int_fullrecoValidationKJpsiee',
        #     mode="train",
        #     max_N_samples_per=-1,
        #     splits=1,
        #     use_weights=False,
        # )
        # dataloader = DataLoader(dataset, batch_size=256, shuffle=False)
        # val_dataloaders["physics_KJpsiee_3"] = dataloader

    branch_options["path"] = training_info["path"]
    branch_options["particles_involved"] = training_info["particles_involved"]

    if verbose:
        for option in branch_options:
            if option != "training_weights" and option != "fully_reco_bool":
                print("\n\n", option, "\n", branch_options[option])

    return dataloaders, val_dataloaders, branch_options
